import json

import click

from gen3.external.nih.dbgap_study_registration import dbgapStudyRegistration
from cdislogging import get_logger

logger = get_logger("__name__")


@click.group()
def nih():
    """Commands for reading from NIH APIs"""
    pass


@nih.group()
def dbgap_study_registration():
    """Commands for interacting with the dbgap study registration api"""
    pass


@dbgap_study_registration.command(name="get-metadata")
@click.argument(
    "studies",
    nargs=-1,
)
def get_child_studies(studies):
    """
    Retrieve the study metadata associated with the provided study names.

    This command fetches the metadata associated with the given study names from the dbGaP Study Registration API.

    NOTE: If no version is provided, then the latest version for a study will be used.

    Args:
        studies (str): A space-separated list of study names in which to get metadata for.
    Example:
        gen3 nih dbgap-study-registration get-metadata phs002793 phs002794 phs002795
        > {
            "phs002793.v2.p1": {
                "@uid": "48490",
                "@whole_study_id": "44467",
                "@phs": "002793",
                "@v": "2",
                "@p": "1",
                "@createDate": "2022-09-14T11:39:44-05:00",
                "@completedByGPADate": "2022-11-07T09:51:24-05:00",
                "@modDate": "2022-12-21T09:03:26-05:00",
                "@maxParentChildStudyModDate": "2022-12-21T09:04:04-05:00",
                "@handle": "SLICE",
                "@num_participants": "500",
                "StudyInfo": {
                    "@accession": "phs002793.v2.p1",
                    "@parentAccession": "phs002793.v2.p1",
                    "childAccession": [
                        "phs002795.v1.p1",
                        "phs002796.v1.p1",
                        "phs002797.v1.p1",
                        "phs002798.v1.p1",
                        "phs002794.v2.p1"
                    ],
                ...truncated for brevity...
            }
        }
    """
    result = dbgapStudyRegistration().get_metadata_for_ids(studies)

    if not result:
        logger.info(f"No study found for {studies}")
    else:
        click.echo(f"{json.dumps(result, indent=4)}")


@dbgap_study_registration.command(name="get-child-studies")
@click.argument(
    "studies",
    nargs=-1,
)
def get_child_studies(studies):
    """
    Retrieve the child studies associated with the provided study names.

    This command fetches the child studies associated with the specified parent study names
    from the dbGaP Study Registration API.

    NOTE: If no version is provided, then the latest version for a study will be used.

    Args:
        studies (str): A space-separated list of parent study names for which to fetch child studies.
    Example:
        gen3 nih dbgap-study-registration get-child-studies phs002793
        > {
              "phs002793.v2.p1": ["phs002795.v1.p1", "phs002796.v1.p1", "phs002797.v1.p1", "phs002798.v1.p1"]
           }
        gen3 nih dbgap-study-registration get-child-studies phs002076 phs002793
         > {
              "phs002076.v2.p1": ["phs002077.v2.p1"],
              "phs002793.v2.p1": ["phs002795.v1.p1", "phs002796.v1.p1", "phs002797.v1.p1", "phs002798.v1.p1"]
            }
    """
    result = dbgapStudyRegistration().get_child_studies_for_ids(studies)

    if not result:
        logger.info(f"No child studies found for {studies}")
    else:
        click.echo(f"{json.dumps(result, indent=4)}")


@dbgap_study_registration.command(name="get-parent-studies")
@click.argument(
    "studies",
    nargs=-1,
)
def get_parent_studies(studies):
    """
    Retrieve the parent study associated with each of the provided study names.

    This command fetches the parent study associated with each of the specified child study names
    from the dbGaP Study Registration API.

    NOTE: If no version is provided, then the latest version for a study will be used.

    Args:
        studies (str): A space-separated list of child study names for which to fetch their parent study.
    Example:
        gen3 nih dbgap-study-registration get-parent-studies phs002795
        > {
              "phs002795.v1.p1": "phs002793.v2.p1"
           }
        gen3 nih dbgap-study-registration get-parent-studies phs002795 phs002796 phs002793
         > {
              "phs002795.v1.p1": "phs002793.v2.p1",
              "phs002796.v1.p1": "phs002793.v2.p1"
              "phs002793.v2.p1": None
            }
    """
    result = dbgapStudyRegistration().get_parent_studies_for_ids(studies)

    if not result:
        logger.info(f"No parent studies found for any {studies}")
    else:
        click.echo(f"{json.dumps(result, indent=4)}")



import logging
import os
from importlib.metadata import version

import click

import cdislogging
import gen3.cli.auth as auth
import gen3.cli.pfb as pfb
import gen3.cli.wss as wss
import gen3.cli.discovery as discovery
import gen3.cli.configure as configure
import gen3.cli.objects as objects
import gen3.cli.file as file
import gen3.cli.drs_pull as drs_pull
import gen3.cli.users as users
import gen3.cli.wrap as wrap
import gen3
from gen3 import logging as sdklogging
from gen3.cli import nih


class AuthFactory:
    def __init__(self, refresh_file):
        self.refresh_file = refresh_file
        self._cache = None

    def get(self):
        """Lazy factory"""
        if self._cache:
            return self._cache
        self._cache = gen3.auth.Gen3Auth(refresh_file=self.refresh_file)
        return self._cache


@click.group()
@click.option(
    "--auth",
    "auth_config",
    default=os.getenv("GEN3_API_KEY", None),
    help="""authentication source, by default expects an API key in "~/.gen3/credentials.json".
    Has special support for token service: "idp://wts/<idp>", and raw access tokens
    "accesstoken:///<token>",
    otherwise a path to an API key or basename of key under ~/.gen3/ can be used.
    Default value is "credentials" if ~/.gen3/credentials.json exists, otherwise "idp://wts/local"
    """,
)
@click.option(
    "--endpoint",
    "endpoint",
    default=os.getenv("GEN3_ENDPOINT", None),
    help="commons hostname - optional if API Key given in `auth`",
)
@click.option(
    "-v",
    "verbose_logs",
    is_flag=True,
    default=False,
    help="verbose logs show INFO, WARNING & ERROR logs",
)
@click.option(
    "-vv",
    "very_verbose_logs",
    is_flag=True,
    default=False,
    help="very verbose logs show DEGUG, INFO, WARNING & ERROR logs",
)
@click.option(
    "--only-error-logs",
    "only_error_logs",
    is_flag=True,
    default=False,
    help="only show ERROR logs",
)
@click.option(
    "--silent",
    "silent",
    is_flag=True,
    default=False,
    help="don't show ANY logs",
)
@click.option(
    "--commons_url",
    "commons_url",
    default=os.getenv("GEN3_COMMONS_URL", None),
    help="commons url for fetching file metadata from if different than endpoint",
)
@click.pass_context
@click.version_option(version=version("gen3"))
def main(
    ctx,
    auth_config,
    endpoint,
    verbose_logs,
    very_verbose_logs,
    only_error_logs,
    silent,
    commons_url,
):
    """Gen3 Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj["auth_config"] = auth_config
    ctx.obj["endpoint"] = endpoint
    ctx.obj["commons_url"] = commons_url
    ctx.obj["auth_factory"] = AuthFactory(auth_config)

    if silent:
        # we still need to define the logger, the log_level here doesn't
        # really matter b/c we immediately disable all logging
        logger = cdislogging.get_logger(
            __name__, format=gen3.LOG_FORMAT, log_level="debug"
        )
        # disables all logging
        logging.disable(logging.CRITICAL)
    elif very_verbose_logs:
        logger = cdislogging.get_logger(
            __name__, format=gen3.LOG_FORMAT, log_level="debug"
        )
        sdklogging.setLevel("DEBUG")
    elif verbose_logs:
        logger = cdislogging.get_logger(
            __name__, format=gen3.LOG_FORMAT, log_level="info"
        )
        sdklogging.setLevel("INFO")
    elif only_error_logs:
        logger = cdislogging.get_logger(
            __name__, format=gen3.LOG_FORMAT, log_level="error"
        )
        sdklogging.setLevel("ERROR")
    else:
        logger = cdislogging.get_logger(
            __name__, format=gen3.LOG_FORMAT, log_level="warning"
        )
        sdklogging.setLevel("WARNING")


main.add_command(auth.auth)
main.add_command(pfb.pfb)
main.add_command(wss.wss)
main.add_command(discovery.discovery)
main.add_command(configure.configure)
main.add_command(objects.objects)
main.add_command(drs_pull.drs_pull)
main.add_command(file.file)
main.add_command(nih.nih)
main.add_command(users.users)
main.add_command(wrap.run)
main()



from cdislogging import get_logger

from gen3.doi import (
    DigitalObjectIdentifierCreator,
    DigitalObjectIdentifierFundingReference,
    DigitalObjectIdentifierDescription,
    DigitalObjectIdentifierTitle,
    DigitalObjectIdenfitierAlternateID,
)
from gen3.external import ExternalMetadataSourceInterface
from gen3.external.nih.dbgap_study_registration import dbgapStudyRegistration
from gen3.external.nih.dbgap_fhir import dbgapFHIR
from gen3.external.nih.utils import get_dbgap_accession_as_parts

logging = get_logger("__name__")


class dbgapDOI(ExternalMetadataSourceInterface):
    """
    Class for interacting with various National Center for
    Biotechnology Information (NCBI)'s database of Genotypes and Phenotypes (dbGaP)
    APIs to get necessary information for a DOI.

    Example Usage:

        ```
        from gen3.external.nih.dbgap_doi import dbgapDOI


        def main():
            studies = [
                "phs000007.v32",
                "phs000179.v1.p1.c1",
            ]

            dbgap_doi = dbgapDOI(publisher="Gen3 Foobar Instance")
            all_doi_data = dbgap_doi.get_metadata_for_ids(ids=studies)

            print(metadata)


        if __name__ == "__main__":
            main()
        ```

    See doi.py module for more information on how to use this DOI metadata to
    actually interact with Datacite's API and mint DOIs.

    """

    def __init__(
        self,
        publisher,
    ):
        """
        Initiatilize the class for interacting with dbgap APIs to get
        DOI information.

        Args:
            publisher (str): The DOI Publisher that should be added as the
                'Publisher' in the DOI Metadata. Per the Datacite docs, this is:
                    "The name of the entity that holds, archives, publishes prints,
                    distributes, releases, issues, or produces the resource. This
                    property will be used to formulate the citation, so consider the
                    prominence of the role. For software, use Publisher for the code
                    repository. If there is an entity other than a code repository,
                    that "holds, archives, publishes, prints, distributes, releases,
                    issues, or produces" the code, use the property
                    Contributor/contributorType/ hostingInstitution for the code
                    repository."
        """
        self.publisher = publisher

    def get_metadata_for_ids(self, ids):
        """
        Return DOI metadata for each of the provided IDs.

        Args:
            ids (List[str]): list of IDs to query for

        Returns:
            Dict[dict]: metadata for each of the provided IDs (which
                        are the keys in the returned dict)
        """
        logging.info("Getting dbGaP DOI metadata...")
        logging.debug(f"Provided ids for dbGaP DOI metadata: {ids}")

        dbgap_study_registration = dbgapStudyRegistration()
        dbgap_fhir = dbgapFHIR()

        dbgap_study_registration_metadata = (
            dbgap_study_registration.get_metadata_for_ids(ids)
        )
        dbgap_fhir_metadata = dbgap_fhir.get_metadata_for_ids(ids)

        all_doi_metadata = {}

        for phsid in ids:
            if not dbgap_fhir_metadata.get(phsid):
                logging.error(
                    f"{phsid} is missing dbGaP FHIR metadata. Cannot "
                    f"continue creating a DOI without it. Skipping..."
                )
                continue

            if not dbgap_study_registration_metadata.get(phsid):
                logging.error(
                    f"{phsid} is missing dbGaP Study Registration "
                    f"metadata. Cannot continue creating a DOI without it. Skipping..."
                )
                continue

            doi_metadata = {}

            # 1) required fields
            doi_metadata["creators"] = dbgapDOI._get_doi_creators(
                phsid, dbgap_study_registration_metadata
            )
            doi_metadata["titles"] = dbgapDOI._get_doi_title(phsid, dbgap_fhir_metadata)
            doi_metadata["publication_year"] = dbgapDOI._get_doi_publication_year(
                phsid, dbgap_fhir_metadata
            )
            doi_metadata["doi_type_general"] = "Dataset"

            # publisher is provided
            doi_metadata["publisher"] = self.publisher

            # NOTE: This does NOT include the required landing page URL
            #       b/c this requires the final ID (which should be generated
            #       elsewhere).
            # doi_metadata["url"] = None

            # 2) optional fields
            doi_metadata["version"] = dbgapDOI._get_doi_version(
                phsid, dbgap_fhir_metadata
            )
            doi_metadata["contributors"] = dbgapDOI._get_doi_contributors(
                phsid, dbgap_study_registration_metadata, dbgap_fhir_metadata
            )
            doi_metadata["descriptions"] = dbgapDOI._get_doi_descriptions(
                phsid, dbgap_fhir_metadata
            )
            doi_metadata[
                "alternateIdentifiers"
            ] = dbgapDOI._get_doi_alternate_identifiers(phsid, dbgap_fhir_metadata)
            doi_metadata["fundingReferences"] = dbgapDOI._get_doi_funding(
                phsid, dbgap_fhir_metadata
            )

            all_doi_metadata[phsid] = doi_metadata

        return all_doi_metadata

    @staticmethod
    def _get_doi_creators(phsid, dbgap_study_registration_metadata):
        result = []

        persons = (
            dbgap_study_registration_metadata.get(phsid, {})
            .get("Authority", {})
            .get("Persons", {})
            .get("Person", [])
        )
        for person in persons:
            # dbgap's API is inconsistent. Sometimes this is represented
            # like this: {'@allow_direct_access': 'true', '#text': 'PI'} and sometimes
            # it's just a string like "PI". This handles both cases to determine
            # the PI
            if person.get("Role") == "PI" or (
                "#text" in person.get("Role")
                and person.get("Role").get("#text") == "PI"
            ):
                creator = DigitalObjectIdentifierCreator(
                    name=person.get("@lname", "") + ", " + person.get("@fname", ""),
                    name_type=DigitalObjectIdentifierCreator.NAME_TYPE_PERSON,
                    given_name=person.get("@fname", ""),
                    family_name=person.get("@lname", ""),
                )
                result.append(creator.as_dict())

        return result

    @staticmethod
    def _get_doi_title(phsid, dbgap_fhir_metadata):
        result = []

        title = DigitalObjectIdentifierTitle(
            dbgap_fhir_metadata.get(phsid, {}).get("Title")
        )

        result.append(title.as_dict())

        return result

    @staticmethod
    def _get_doi_publication_year(phsid, dbgap_fhir_metadata):
        date = dbgap_fhir_metadata.get(phsid, {}).get("ReleaseDate")

        if not date:
            logging.debug(f"dbgap_fhir_metadata: {dbgap_fhir_metadata}")
            raise Exception(
                f"ReleaseDate from dbgap FHIR does not match expected pattern "
                f"YYYY-MM-DD: '{date}'. Unable to parse."
            )

        if date:
            date = date.split("-")[0]

            if len(date) != 4:
                logging.debug(f"dbgap_fhir_metadata: {dbgap_fhir_metadata}")
                raise Exception(
                    f"ReleaseDate from dbgap FHIR does not match expected pattern "
                    f"YYYY-MM-DD: '{date}'. Unable to parse."
                )

        return date

    @staticmethod
    def _get_doi_contributors(
        phsid, dbgap_study_registration_metadata, dbgap_fhir_metadata
    ):
        result = []

        persons = (
            dbgap_study_registration_metadata.get(phsid, {})
            .get("Authority", {})
            .get("Persons", {})
            .get("Person", [])
        )
        for person in persons:
            # dbgap's API is inconsistent. Sometimes this is represented
            # like this: {'@allow_direct_access': 'true', '#text': 'PI'} and sometimes
            # it's just a string like "PI". This handles both cases to determine
            # the PI and PI_ASSIST persons
            if (
                person.get("Role") == "PI"
                or person.get("Role") == "PI_ASSIST"
                or (
                    "#text" in person.get("Role")
                    and (
                        person.get("Role").get("#text") == "PI"
                        or person.get("Role").get("#text") == "PI_ASSIST"
                    )
                )
            ):
                creator = DigitalObjectIdentifierCreator(
                    name=person.get("@lname", "") + ", " + person.get("@fname", ""),
                    name_type=DigitalObjectIdentifierCreator.NAME_TYPE_PERSON,
                    given_name=person.get("@fname", ""),
                    family_name=person.get("@lname", ""),
                )
                result.append(creator.as_dict())

                if person.get("Organization"):
                    creator = DigitalObjectIdentifierCreator(
                        name=person.get("Organization"),
                        name_type=DigitalObjectIdentifierCreator.NAME_TYPE_ORGANIZATION,
                    )
                    result.append(creator.as_dict())

        return result

    @staticmethod
    def _get_doi_version(phsid, dbgap_fhir_metadata):
        version = 1
        ids = dbgap_fhir_metadata.get(phsid, {}).get("Identifier")
        for phsid in ids:
            standardized_phsid_parts = get_dbgap_accession_as_parts(phsid)

            # if there is a phsid parsed, then this Identifier matches
            # the pattern of the full phsid, so we can parse the version out
            if standardized_phsid_parts["phsid"]:
                version = standardized_phsid_parts["version_number"]

        return version

    @staticmethod
    def _get_doi_descriptions(phsid, dbgap_fhir_metadata):
        result = []

        description = DigitalObjectIdentifierDescription(
            description=dbgap_fhir_metadata.get(phsid, {}).get("Description")
        )
        result.append(description.as_dict())

        return result

    @staticmethod
    def _get_doi_alternate_identifiers(phsid, dbgap_fhir_metadata):
        result = []

        for alternate_id in dbgap_fhir_metadata.get(phsid, {}).get("Identifier"):
            alternate_id_obj = DigitalObjectIdenfitierAlternateID(
                alternate_id=alternate_id, alternate_id_type="dbGaP"
            )
            result.append(alternate_id_obj.as_dict())

        return result

    @staticmethod
    def _get_doi_funding(phsid, dbgap_fhir_metadata):
        return DigitalObjectIdentifierFundingReference(
            name=dbgap_fhir_metadata.get(phsid, {}).get("Sponsor")
        ).as_dict()



import csv
import json
from cdislogging import get_logger
import os
import time
import sys
import traceback
import re

from drsclient.client import DrsClient
from gen3.auth import Gen3Auth
from gen3.utils import UUID_FORMAT, SIZE_FORMAT, _verify_format, _standardize_str
from gen3.tools.utils import (
    GUID_COLUMN_NAMES,
    SIZE_COLUMN_NAMES,
    BUNDLENAME_COLUMN_NAME,
    IDS_COLUMN_NAME,
    DESCRIPTION_COLUMN_NAME,
    CHECKSUMS_COLUMN_NAME,
    TYPE_COLUMN_NAME,
    ALIASES_COLUMN_NAME,
)

logging = get_logger("__name__")

"""
NOTES:
- Assumption is that the user creates the manifest with the correct hierarchy.

Bundle manifest format

bundle_name, ids, bundle_guid (optional) // can also include other optional fields
Bundle-B, [File-1-GUID, File-2-GUID],
Bundle-C, [File-3-GUID, File-4-GUID],
Bundle-A, [Bundle-B, Bundle-C],

json representaion:
[
  {
    "Bundle_name": "Bundle-B",
    "Ids": [ "File-1-GUID", "File-2-GUID" ],
     // optional fields can also be added
  },
  {
     "Bundle_name": "Bundle-C"
     "Ids": [ "File-3-GUID", "File-4-GUID"]
  },
  {
     "Bundle_name": "Bundle-A"
     "Ids": [ "Bundle-B, "Bundle-C"]
  }
]
"""

ACCEPTABLE_HASHES = {
    "md5": re.compile(r"^[0-9a-f]{32}$").match,
    "sha1": re.compile(r"^[0-9a-f]{40}$").match,
    "sha256": re.compile(r"^[0-9a-f]{64}$").match,
    "sha512": re.compile(r"^[0-9a-f]{128}$").match,
    "crc": re.compile(r"^[0-9a-f]{8}$").match,
    "etag": re.compile(r"^[0-9a-f]{32}(-\d+)?$").match,
}


def validate_hashes(hashes):
    """
    Validate hashes against known and valid hashing algorithms.
    """
    if not all(h in ACCEPTABLE_HASHES for h in hashes):
        return False
    if not all(ACCEPTABLE_HASHES[h](v) for h, v in hashes.items()):
        return False
    return True


def _replace_bundle_name_with_guid(bundles, bundle_name_to_guid):
    """
    replace the bundle names in the list of bundles with its associated guid
    """
    new_list = []
    for bundle in bundles:
        if _verify_format(bundle, UUID_FORMAT):
            new_list.append(bundle)
        else:
            if bundle in bundle_name_to_guid:
                new_list.append(bundle_name_to_guid[bundle])
    return new_list


def _join_type_and_checksum(type_list, checksum_list):
    """
    Join checksum and their correlated type together to the following format:
    "checksums": [{"type":"md5", "checksum":"abcdefg}, {"type":"sha256", "checksum":"abcd12345"}] # pragma: allowlist secret
    """
    checksums = [
        {
            "type": c_type,
            "checksum": checksum,
        }
        for c_type, checksum in zip(type_list, checksum_list)
    ]
    return checksums


def _verify_and_process_bundle_manifest(manifest_file, manifest_file_delimiter="\t"):
    """
    Verify the manifest and create list of all records that needs to be ingested

    Args:
        manifest_file(str): the path to the input manifest
        manifest_file_delimiter(str): delimiter

    Returns:
        records(list): List of all records to be ingested
        bundle_name_to_guid(dict): dict containing bundle_name to bundle_guid association
    NOTE: we dont do a check if indexd contains these ids since it checks that in POST /bundle
    """
    pass_verification = True
    records = []
    bundle_name_to_guid = {}

    with open(manifest_file, encoding="utf-8-sig") as manifest:
        reader = csv.DictReader(manifest, delimiter=manifest_file_delimiter)
        for row_n, row in enumerate(reader, 1):
            record = {}
            for key, value in row.items():
                if key in BUNDLENAME_COLUMN_NAME:
                    # make sure bundle_name exists. bundle_name is a required field.
                    if value:
                        bundle_name = value
                        record["name"] = bundle_name
                        if bundle_name not in bundle_name_to_guid:
                            bundle_name_to_guid[
                                bundle_name
                            ] = ""  # to keep track of the available bundles.
                        else:
                            logging.error(
                                "ERROR: bundle_name {} at row {} is not unique".format(
                                    bundle_name, row_n
                                )
                            )
                            pass_verification = False
                    else:
                        logging.error(
                            "ERROR: row {} does not contain bundle_name. bundle_name is required".format(
                                row_n
                            )
                        )
                        pass_verification = False
                elif key in IDS_COLUMN_NAME:
                    standard = (
                        _standardize_str(value)
                        .strip()
                        .lstrip("[")
                        .rstrip("]")
                        .split(" ")
                    )
                    item_ids = []
                    for item in standard:
                        item_id = item.replace("'", "").replace('"', "")
                        if item_id == record["name"]:
                            logging.error(
                                "Error: Bundle at row {} contains itself".format(row_n)
                            )
                            pass_verification = False
                        elif (
                            _verify_format(item_id, UUID_FORMAT)
                            or item_id in bundle_name_to_guid
                        ):  # so if its a bundle name thats in the dictionary then keep it for processing it later
                            item_ids.append(item_id)
                        else:
                            logging.error(
                                "ERROR: {} {} at row {} must either be a UUID or reference another bundle in this manifest".format(
                                    key, bundle, row_n
                                )
                            )
                            pass_verification = False
                    record["bundles"] = item_ids
                elif key in GUID_COLUMN_NAMES and value:
                    if _verify_format(value, UUID_FORMAT):
                        record["guid"] = value
                    else:
                        logging.error(
                            "ERROR: {} {} at row {} is in an incorrect format".format(
                                key, value, row_n
                            )
                        )
                elif (
                    key in TYPE_COLUMN_NAME
                    or key in CHECKSUMS_COLUMN_NAME
                    or key in ALIASES_COLUMN_NAME
                ):
                    if not value:
                        continue
                    standard = (
                        _standardize_str(value)
                        .strip()
                        .lstrip("[")
                        .rstrip("]")
                        .split(" ")
                    )
                    values = []
                    for element in standard:
                        v = element.strip().replace("'", "").replace('"', "")
                        if v:
                            values.append(v)
                    k = "aliases"
                    if key in CHECKSUMS_COLUMN_NAME:
                        k = "checksum"
                    if key in TYPE_COLUMN_NAME:
                        k = "type"
                    record[k] = values
                elif value:
                    if key in SIZE_COLUMN_NAMES:
                        value = int(value)
                    record[key] = value
            if "checksum" in record and "type" in record:
                if len(record["type"]) != len(record["checksum"]):
                    logging.error(
                        "ERROR: Number of checksum types ({}) and checksums ({}) does not match for row {}.".format(
                            len(record["type"]), len(record["checksum"]), row_n
                        )
                    )
                    pass_verification = False
                elif not validate_hashes(dict(zip(record["type"], record["checksum"]))):
                    logging.error("ERROR: invalid checksums at row {}".format(row_n))
                    pass_verification = False
                else:
                    record["checksums"] = _join_type_and_checksum(
                        record["type"], record["checksum"]
                    )
                    del record["type"]
                    del record["checksum"]
            records.append(record)
    if not pass_verification:
        logging.error("The manifest is not in the correct format!")
        return None, None
    return records, bundle_name_to_guid


def _write_csv(records, filename="output_manifest.csv"):
    fieldnames = [
        "name",
        "bundles",
        "guid",
        "size",
        "checksums",
        "description",
        "version",
        "aliases",
    ]
    with open(filename, mode="w") as outfile:
        writer = csv.DictWriter(outfile, delimiter=",", fieldnames=fieldnames)
        writer.writeheader()
        for r in records:
            writer.writerow(r)

    return filename


def ingest_bundle_manifest(
    commons_url,
    manifest_file,
    out_manifest_file="ingest_out.csv",
    manifest_file_delimiter=None,
    auth=None,
    token=None,
):
    logging.info("Starting the process ....")
    start_time = time.perf_counter()
    logging.info("start time: {}".format(start_time))
    commons_url = commons_url.strip("/")

    if not manifest_file_delimiter:
        file_ext = os.path.splitext(manifest_file)
        if file_ext[-1].lower() == ".tsv":
            manifest_file_delimiter = "\t"
        else:
            # default, assume CSV
            manifest_file_delimiter = ","

    try:
        records, bundle_name_to_guid = _verify_and_process_bundle_manifest(
            manifest_file, manifest_file_delimiter
        )
    except Exception as e:
        exc_info = sys.exc_info()
        traceback.print_exception(*exc_info)
        logging.error("Can not read {}. Detail {}".format(manifest_file, e))
        return None

    if not records:
        return None

    drsclient = DrsClient(commons_url, auth=auth, token=token)
    total = len(records)
    # Iterate through the records list and post to indexd
    for start, record in enumerate(records, 1):
        bundle_name = record["name"]

        logging.info("Posting bundle {} of {}".format(start, total))
        # Check the bundle list to make sure they're all guids
        record["bundles"] = _replace_bundle_name_with_guid(
            record["bundles"], bundle_name_to_guid
        )
        resp = drsclient.create(**record)

        if resp.status_code != 200:
            logging.error(
                "Failed to create bundle {}. Status code: {} Detail: {}".format(
                    bundle_name, resp.status_code, resp.text
                )
            )
        else:
            rec = resp.json()
            guid = rec["bundle_id"]
            logging.info("Successfully created bundle {}".format(bundle_name))
            # Associate bundle_name to its guid created by indexd
            bundle_name_to_guid[bundle_name] = guid
            if "guid" not in record:
                record["guid"] = guid

    logging.info("Published all {} bundles".format(total))
    logging.info("Start writng output manifest . . .")

    _write_csv(records)

    logging.info("Output csv created")
    logging.info("Done!")

    return records



from cdislogging import get_logger
import click
import gen3.configure as config_tool

logging = get_logger("__name__")


@click.command()
@click.option("--profile", help="name of the profile to name for this credentials")
@click.option("--cred", help="path to the credentials.json")
def configure(profile, cred):
    """[unfinished] Commands to configure multiple profiles with corresponding credentials

    ./gen3 configure --profile=<profile-name> --cred=<path-to-credential/cred.json>
    """

    logging.info(f"Configuring profile [ {profile} ] with credentials at {cred}")

    try:
        profile_title, new_lines = config_tool.get_profile_from_creds(profile, cred)
        lines = config_tool.get_current_config_lines()
        config_tool.update_config_lines(lines, profile_title, new_lines)
    except Exception as e:
        logging.warning(str(e))
        raise e



from typing import List, Optional
import requests
import json
from cdislogging import get_logger
import os
import inspect
from datetime import datetime, timezone, timedelta
from pathlib import Path

# DRS Compact identifiers are resolved to hostnames using a set of resolvers
# NOTE: that the policy is to minimize the requests to dataguids.org or other DRS
# resolvers, which is why the resolution order start local and expands outward.

# Environment Variables for controlling how DRS hostname are resolved and
# cached, if DRS should be cached, and the order to execute the resolvers
DRS_CACHE_EXPIRE_DURATION = os.getenv("DRS_CACHE_EXPIRE_DURATION", 2)  # In Days
DRS_RESOLVER_HOSTNAME = os.getenv("DRS_RESOLVER_HOSTNAME", "https://dataguids.org")
DRS_CACHE_EXPIRE = timedelta(days=DRS_CACHE_EXPIRE_DURATION)
LOCALLY_CACHE_RESOLVED_HOSTS = os.getenv("LOCALLY_CACHE_RESOLVED_HOSTS", True)
DRS_RESOLUTION_ORDER = os.getenv(
    "DRS_RESOLUTION_ORDER", "cache_file:commons_mds:dataguids_dist:dataguids"
)

DRS_CACHE = os.getenv(
    "DRS_CACHE", str(Path(Path.home(), ".drs_cache", "resolved_drs_hosts.json"))
)

logger = get_logger("download", log_level="warning")


def clean_dist_entry(s: str) -> str:
    """
    Cleans the string returning a proper DRS prefix
           s: string to clean
    @return: cleaned string
    """
    return s.replace("\\.", ".").replace(".*", "")


def clean_http_url(s: str) -> str:
    """
    Cleans input string removing http(s) prefix and all trailing paths
           s: string to clean
    @return: cleaned string
    """
    return (
        s.replace("/index", "")[::-1]
        .replace("/", "", 1)[::-1]
        .replace("http://", "")
        .replace("https://", "")
        .replace("/ga4gh/drs/v1/objects", "")
    )


def create_local_drs_cache(data: dict, cache_path: str = None) -> bool:
    """
    Creates a file caching resolved DRS hosts passed in the data paramter. Will logged error if unable
    to create (typically a bad path or invalid permissions)
           data: dictionary of resolved DRS hosts
           cache_path: path to store the cache file. If not set will use the environment variable DRS_CACHE
    @return: True if successfully written
    """
    if cache_path is None:
        cache_path = DRS_CACHE
    try:
        cache_path = Path(cache_path)
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        # create timestamp for when the cache file was created
        with open(cache_path, "wt") as fout:
            json.dump(
                {
                    "info": {
                        "created": datetime.now(timezone.utc).strftime(
                            "%m/%d/%Y %H:%M:%S:%z"
                        )
                    },
                    "cache": data,
                },
                fout,
            )
        return True

    except IOError as ex:
        logger.critical(f"cannot open {cache_path}: {ex}")
    return False


def append_to_local_drs_cache(data: dict, cache_path: str = None) -> bool:
    """
    Appends/replaces a resolved DRS hostname in the local cache. If no cache file
    exists, it will create one.
           data: DRS resolution object to add
           cache_path: path to local cache
    @return: true if successful
    """
    if cache_path is None:
        cache_path = DRS_CACHE
    try:
        cache_path = Path(cache_path)
        if cache_path.exists() is False:  # no cache exists so create a cache
            return create_local_drs_cache(data, cache_path)

        with open(cache_path, "rt") as fin:
            cache_data = json.load(fin)
            cache_data["cache"] = {**cache_data["cache"], **data}
            with open(cache_path, "wt") as fout:
                json.dump(cache_data, fout)

        return True

    except IOError as ex:
        logger.critical(f"cannot open {cache_path}: {ex}")
    except json.JSONDecodeError as ex:
        logger.error(f"json file cannot be parsed: {ex}")
    return False


def resolve_drs_from_local_cache(
    identifier: str, _: str = None, **kwargs
) -> Optional[str]:
    """Resolves a compact DRS prefix by reading a local cache file

    The format of the cache file is JSON.
    If the id is found and considered "fresh" return the identifier, otherwise
    return None indicating that the id is not resolved. Resolved hostnames are considered
    fresh if the entry's timestamp is less than the expire value, which is in days.

    Args
           identifier: DRS prefix to resolve
           _: unused
           kwargs: optional parameters:
                "cache_dir": directory to store the cache in
           Returns a hostname if resolved, otherwise None
    """
    filename = kwargs.get("cache_dir", DRS_CACHE)
    if filename is None:
        return None
    # if no cache file, then return None, as this is not really an error
    if Path(DRS_CACHE).exists() is False:
        return None
    try:
        with open(filename, "rt") as fin:
            data = json.load(fin)
            if identifier in data["cache"]:
                timestamp = datetime.strptime(
                    data["cache"][identifier].get("created", "1/1/1900 00:00:01:+0000"),
                    "%m/%d/%Y %H:%M:%S:%z",
                )
                if (datetime.now(timezone.utc) - timestamp) > DRS_CACHE_EXPIRE:
                    fin.close()  # cache is expired return and one of the other resolvers will recreate it
                    return None
                return data["cache"][identifier].get("host", None)

    except json.JSONDecodeError as ex:
        logger.error(f"json cache file cannot be parsed: {ex}")

    return None


def resolve_compact_drs_using_indexd_dist(
    identifier: str,
    cache_results: bool = LOCALLY_CACHE_RESOLVED_HOSTS,
    resolver_hostname: str = DRS_RESOLVER_HOSTNAME,
):
    """Resolves a compact DRS prefix by by performing a lookup via a commons indexd/_dist endpoint.

    If the id is found it, optionally caches the results, and returns the identifier, otherwise
    return None indicating that the id is not resolved.

    Args
           identifier: DRS prefix to resolve
           cache_results: set to true to write local cache file
           resolver_hostname: url of indexd server to query. This is typically dataguids.org
           Returns a hostname if resolved, otherwise None
    """
    try:
        response = requests.get(f"{resolver_hostname}/index/_dist")
        response.raise_for_status()
        results = response.json()

        # convert to cached format
        data = {}
        for entry in results:
            if entry["type"] != "indexd":
                continue
            host = clean_http_url(entry["host"])
            name = entry.get("name", "")
            for x in entry["hints"]:
                entry_id = clean_dist_entry(x)
                data[entry_id] = {
                    "host": host,
                    "name": name,
                    "type": entry["type"],
                    "created": datetime.now(timezone.utc).strftime(
                        "%m/%d/%Y %H:%M:%S:%z"
                    ),
                }

        if identifier in data:
            if cache_results:  # write the results to cache since we have a
                # number of results, caching them will potentially
                # save lookup time in the future
                append_to_local_drs_cache(data)

            return data[identifier]["host"]

    except requests.exceptions.HTTPError as exc:
        logger.critical(
            f"HTTP Error accessing {DRS_RESOLVER_HOSTNAME}: {exc.response.status_code}"
        )
    return None


def resolve_drs_using_metadata_service(
    identifier: str,
    metadata_service_url: str,
    cache_results: bool = LOCALLY_CACHE_RESOLVED_HOSTS,
) -> Optional[str]:
    """Resolves a compact DRS prefix by by performing a lookup into a Gen3 AggregateMDS.

    If the id is found it, optionally caches the results, and returns the identifier, otherwise
    return None indicating that the id is not resolved.

    Args
           identifier: DRS prefix to resolve
           metadata_service_url: url of MDS server to query.
           cache_results: set to true to write local cache file
           Returns a hostname if resolved, otherwise None
    """
    if not metadata_service_url:
        return None
    try:
        response = requests.get(f"{metadata_service_url}/{identifier}")
        response.raise_for_status()
        results = response.json()
        if "host" in results:
            hn = clean_http_url(results["host"])
            if cache_results:
                data = {
                    "identifier": {
                        "host": hn,
                        "name": results.get("name", ""),
                        "type": "indexd",
                        "created": datetime.now(timezone.utc).strftime(
                            "%m/%d/%Y %H:%M:%S:%z"
                        ),
                    }
                }
                append_to_local_drs_cache(data)
            return hn

    except requests.exceptions.HTTPError as exc:
        if exc.response.status_code != 404:
            logger.info(
                f"HTTP Error accessing {DRS_RESOLVER_HOSTNAME}: {exc.response.status_code}"
            )
        return None


def resolve_compact_drs_using_official_resolver(
    identifier: str,
    object_id: str,
    cache_results: bool = LOCALLY_CACHE_RESOLVED_HOSTS,
    resolver_hostname: str = DRS_RESOLVER_HOSTNAME,
) -> Optional[str]:
    """Resolves a DRS prefix or objectId using a DRS resolver

    Use DRS_RESOLVER_HOSTNAME (default: dataguids.org) to resolve identifier
    At this time there are two possible ways to resolve a compact ID
    query https://dataguids.org/index/ with the object id and
    look for the hostname in the field "from_index_service"
    failing that try to access dataguids metadata service

    Args
           identifier: DRS prefix to resolve
           object_id: objectid with DRS prefix to resolve
           cache_results: set to true to write local cache file
           resolver_hostname: url of indexd server to query. This is typically dataguids.org

           Returns hostname if resolved, otherwise None
    """

    try:
        response = requests.get(f"{resolver_hostname}/index/{object_id}")
        response.raise_for_status()
        results = response.json()
        hn = results.get("from_index_service", {}).get("host", None)
        if (hn) is not None:
            hn = clean_http_url(hn)
            if cache_results:
                # create and entry to append to the cache
                data = {
                    "identifier": {
                        "host": hn,
                        "name": results.get("from_index_service", {}).get("name", None),
                        "type": "indexd",
                        "created": datetime.now(timezone.utc).strftime(
                            "%m/%d/%Y %H:%M:%S:%z"
                        ),
                    }
                }
                append_to_local_drs_cache(data)

            return hn

    except requests.exceptions.HTTPError as exc:
        if exc.response.status_code == 404:
            # not found try using the MDS
            pass
        else:
            return None

    return resolve_drs_using_metadata_service(
        identifier, f"{resolver_hostname}/mds/metadata", cache_results
    )


def resolve_drs_using_commons_mds(
    identifier: str,
    _: str,
    metadata_service_url: str,
    cache_results: bool = LOCALLY_CACHE_RESOLVED_HOSTS,
) -> Optional[str]:
    """Wrapper to use the resolve_drs_using_metadata_service as a DRS resolver functions

    Args:
           identifier: DRS prefix to resolve
           _: unused
           metadata_service_url: url of MDS server to query.
           cache_results: set to true to write local cache file
           Returns a hostname if resolved, otherwise None
    """
    return resolve_drs_using_metadata_service(
        identifier, metadata_service_url, cache_results
    )


# TODO: provide methods to turn this into a plugin architecture
REGISTERED_DRS_RESOLVERS = {
    "cache_file": resolve_drs_from_local_cache,
    "commons_mds": resolve_drs_using_commons_mds,
    "dataguids_dist": resolve_compact_drs_using_indexd_dist,
    "dataguids": resolve_compact_drs_using_official_resolver,
    # TODO "identifiers.org"
}


def resolve_drs_via_list(
    resolvers_to_try: List[str], identifier, object_id, **kwargs
) -> Optional[str]:
    """Attempts to resolve DRS host by iterating through a list of methods
    Args
       resolvers_to_try: List of  DRS resolvers to execute.
       identifier: DRS prefix to resolve
       object_id: or objectid with DRS prefix to resolve
       kwargs: per resolver options
    Returns hostname if resolved, None otherwise. This would indicate that
    the DRS identifier cannot be resolved.
    """
    tried = []
    for how in resolvers_to_try:
        tried.append(how)
        resolver = REGISTERED_DRS_RESOLVERS.get(how, None)
        if resolver is None:
            continue
        sig = inspect.signature(resolver)
        filter_keys = [
            param.name
            for param in sig.parameters.values()
            if param.kind == param.POSITIONAL_OR_KEYWORD
            and param.name not in ["identifier", "object_id", "_"]
        ]
        parameters_dict = {
            filter_key: kwargs[filter_key]
            for filter_key in filter_keys
            if filter_key in kwargs
        }

        host = resolver(identifier, object_id, **parameters_dict)
        if host is not None:
            logger.info(f"resolved {identifier} tried {tried}")
            return host

    logger.warning(f"unable to resolve {identifier} or {object_id}, tried {tried}")
    return None


def resolve_drs(identifier, object_id, **kwargs):
    """Resolve DRS identifier or object id using the various registered DRS resolvers

    This is top level call used by the DRS downloader

    Args
       identifier: DRS prefix to resolve
       object_id: or objectid with DRS prefix to resolve
       kwargs: per resolver options

    Examples:

        >>> resolve_drs("dg.4DFC", None)
    """
    return resolve_drs_via_list(
        DRS_RESOLUTION_ORDER.split(":"), identifier, object_id, **kwargs
    )



"""
Contains class for interacting with Gen3's Job Dispatching Service(s).
"""
import aiohttp
import asyncio
import backoff
import json
import requests
import urllib.parse
from cdislogging import get_logger

import sys
import time

from gen3.utils import (
    append_query_params,
    DEFAULT_BACKOFF_SETTINGS,
    raise_for_status_and_print_error,
)

# sower's "action" mapping to the relevant job
INGEST_METADATA_JOB = "ingest-metadata-manifest"
DBGAP_METADATA_JOB = "get-dbgap-metadata"
INDEX_MANIFEST_JOB = "index-object-manifest"
DOWNLOAD_MANIFEST_JOB = "download-indexd-manifest"
MERGE_MANIFEST_JOB = "merge-manifests"

logging = get_logger("__name__")


class Gen3Jobs:
    """
    A class for interacting with the Gen3's Job Dispatching Service(s).

    Examples:
        This generates the Gen3Jobs class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... jobs = Gen3Jobs(auth)
    """

    def __init__(self, endpoint=None, auth_provider=None, service_location="job"):
        """
        Initialization for instance of the class to setup basic endpoint info.

        Args:
            auth_provider (Gen3Auth, optional): Gen3Auth class to handle passing your
                token, required for admin endpoints
            service_location (str, optional): deployment location relative to the
                endpoint provided
        """
        # auth_provider legacy interface required endpoint as 1st arg
        auth_provider = auth_provider or endpoint
        endpoint = auth_provider.endpoint.strip("/")
        # if running locally, mds is deployed by itself without a location relative
        # to the commons
        if "http://localhost" in endpoint:
            service_location = ""

        if not endpoint.endswith(service_location):
            endpoint += "/" + service_location

        self.endpoint = endpoint.rstrip("/")
        self._auth_provider = auth_provider

    async def async_run_job_and_wait(self, job_name, job_input, _ssl=None, **kwargs):
        """
        Asynchronous function to create a job, wait for output, and return. Will
        sleep in a linear delay until the job is done, starting with 1 second.

        Args:
            _ssl (None, optional): whether or not to use ssl
            job_name (str): name for the job, can use globals in this file
            job_input (Dict): dictionary of input for the job

        Returns:
            Dict: Response from the endpoint
        """
        job_create_response = await self.async_create_job(job_name, job_input)

        status = {"status": "Running"}
        sleep_time = 3
        while status.get("status") == "Running":
            logging.info(f"job still running, waiting for {sleep_time} seconds...")
            time.sleep(sleep_time)
            sleep_time *= 1.5
            status = await self.async_get_status(job_create_response.get("uid"))
            logging.info(f"{status}")

        logging.info(f"Job is finished!")

        if status.get("status") != "Completed":
            raise Exception(f"Job status not complete: {status.get('status')}.")

        response = await self.async_get_output(job_create_response.get("uid"))
        return response

    def is_healthy(self):
        """
        Return if is healthy or not

        Returns:
            bool: True if healthy
        """
        try:
            response = requests.get(
                self.endpoint + "/_status", auth=self._auth_provider
            )
            raise_for_status_and_print_error(response)
        except Exception as exc:
            logging.error(exc)
            return False

        return response.text == "Healthy"

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_version(self):
        """
        Return the version

        Returns:
            str: the version
        """
        response = requests.get(self.endpoint + "/_version", auth=self._auth_provider)
        raise_for_status_and_print_error(response)
        return response.json().get("version")

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def list_jobs(self):
        """
        List all jobs
        """
        response = requests.get(self.endpoint + "/list", auth=self._auth_provider)
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_job(self, job_name, job_input):
        """
        Create a job with given name and input

        Args:
            job_name (str): name for the job, can use globals in this file
            job_input (Dict): dictionary of input for the job

        Returns:
            Dict: Response from the endpoint
        """
        data = {"action": job_name, "input": job_input}
        response = requests.post(
            self.endpoint + "/dispatch", json=data, auth=self._auth_provider
        )
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_create_job(self, job_name, job_input, _ssl=None, **kwargs):
        async with aiohttp.ClientSession() as session:
            url = self.endpoint + f"/dispatch"
            url_with_params = append_query_params(url, **kwargs)

            data = json.dumps({"action": job_name, "input": job_input})

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            async with session.post(
                url_with_params, data=data, headers=headers, ssl=_ssl
            ) as response:
                raise_for_status_and_print_error(response)
                response = await response.json(content_type=None)
                return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_status(self, job_id):
        """
        Get the status of a previously created job
        """
        response = requests.get(
            self.endpoint + f"/status?UID={job_id}", auth=self._auth_provider
        )
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_status(self, job_id, _ssl=None, **kwargs):
        async with aiohttp.ClientSession() as session:
            url = self.endpoint + f"/status?UID={job_id}"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            async with session.get(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                raise_for_status_and_print_error(response)
                response = await response.json(content_type=None)
                return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_output(self, job_id):
        """
        Get the output of a previously completed job
        """
        response = requests.get(
            self.endpoint + f"/output?UID={job_id}", auth=self._auth_provider
        )
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_output(self, job_id, _ssl=None, **kwargs):
        async with aiohttp.ClientSession() as session:
            url = self.endpoint + f"/output?UID={job_id}"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            async with session.get(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                raise_for_status_and_print_error(response)
                response = await response.json(content_type=None)
                return response



import requests

from gen3.utils import raise_for_status_and_print_error


class Gen3Query:
    """
    Query ElasticSearch data from a Gen3 system.

    Args:
        auth_provider (Gen3Auth): A Gen3Auth class instance.

    Examples:
        This generates the Gen3Query class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(endpoint, refresh_file="credentials.json")
        ... query = Gen3Query(auth)
    """

    def __init__(self, auth_provider):
        self._auth_provider = auth_provider

    def query(
        self,
        data_type,
        fields,
        first=None,
        offset=None,
        filters=None,
        filter_object=None,
        sort_object=None,
        accessibility=None,
        verbose=True,
    ):
        """
        Execute a query against a Data Commons.

        Args:
            data_type (str): Data type to query.
            fields (list): List of fields to return.
            first (int, optional): Number of rows to return (default: 10).
            offset (int, optional): Starting position (default: 0).
            filters: (object, optional): { field: sort method } object. Will filter data with ALL fields EQUAL to the provided respective value. If more complex filters are needed, use the `filter_object` parameter instead.
            filter_object (object, optional): Filter to apply. For syntax details, see https://github.com/uc-cdis/guppy/blob/master/doc/queries.md#filter.
            sort_object (object, optional): { field: sort method } object.
            accessibility (list, optional): One of ["accessible" (default), "unaccessible", "all"]. Only valid when querying a data type in "regular" tier access mode.

        Returns:
            Object: {"data": {<data_type>: [<record>, <record>, ...]}}

        Examples:
            >>> Gen3Query.query(
                data_type="subject",
                first=50,
                fields=[
                    "vital_status",
                    "submitter_id",
                ],
                filters={"vital_status": "Alive"},
                sort_object={"submitter_id": "asc"},
            )
        """
        if not first:
            first = 10
        if not offset:
            offset = 0
        if not sort_object:
            sort_object = {}
        if not accessibility:
            accessibility = "accessible"
        if filters and filter_object:
            raise Exception(
                "Only one of `filters` and `filter_object` can be used at a time."
            )
        if filters:
            filter_object = {
                "AND": [{"=": {field: val}} for field, val in filters.items()]
            }

        if first + offset > 10000:  # ElasticSearch limitation
            sort_fields = [{field: val} for field, val in sort_object.items()]
            data = self.raw_data_download(
                data_type=data_type,
                fields=fields,
                filter_object=filter_object,
                sort_fields=sort_fields,
                accessibility=accessibility,
                first=first,
                offset=offset,
            )
            return {"data": {data_type: data}}

        # convert sort_object to graphql: [ { field_name: "sort_method" } ]
        sorts = [f'{{{field}: "{val}"}}' for field, val in sort_object.items()]
        sort_string = f'[{", ".join(sorts)}]'

        query_string = f"""query($filter: JSON) {{
            {data_type}(
                first: {first},
                offset: {offset},
                sort: {sort_string},
                accessibility: {accessibility},
                filter: $filter
            ) {{
                {" ".join(fields)}
            }}
        }}"""
        variables = {"filter": filter_object}
        return self.graphql_query(query_string=query_string, variables=variables)

    def graphql_query(self, query_string, variables=None):
        """
        Execute a GraphQL query against a Data Commons.

        Args:
            query_txt (str): GraphQL query as text. For syntax details, see https://github.com/uc-cdis/guppy/blob/master/doc/queries.md.
            variables (:obj:`object`, optional): Dictionary of variables to pass with the query.

        Returns:
            Object: {"data": {<data_type>: [<record>, <record>, ...]}}

        Examples:
            >>> query_string = "{ my_index { my_field } }"
            ... Gen3Query.graphql_query(query_string)
        """
        url = f"{self._auth_provider.endpoint}/guppy/graphql"
        response = requests.post(
            url,
            json={"query": query_string, "variables": variables},
            auth=self._auth_provider,
        )
        try:
            raise_for_status_and_print_error(response)
        except Exception:
            print(
                f"Unable to query.\nQuery: {query_string}\nVariables: {variables}\n{response.text}"
            )
            raise
        try:
            return response.json()
        except Exception:
            print(f"Did not receive JSON: {response.text}")
            raise

    def raw_data_download(
        self,
        data_type,
        fields,
        filter_object=None,
        sort_fields=None,
        accessibility=None,
        first=None,
        offset=None,
    ):
        """
        Execute a raw data download against a Data Commons.

        Args:
            data_type (str): Data type to download from.
            fields (list): List of fields to return.
            filter_object (object, optional): Filter to apply. For syntax details, see https://github.com/uc-cdis/guppy/blob/master/doc/queries.md#filter.
            sort_fields (list, optional): List of { field: sort method } objects.
            accessibility (list, optional): One of ["accessible" (default), "unaccessible", "all"]. Only valid when downloading from a data type in "regular" tier access mode.
            first (int, optional): Number of rows to return (default: all rows).
            offset (int, optional): Starting position (default: 0).

        Returns:
            List: [<record>, <record>, ...]

        Examples:
            >>> Gen3Query.raw_data_download(
                    data_type="subject",
                    fields=[
                        "vital_status",
                        "submitter_id",
                        "project_id"
                    ],
                    filter_object={"=": {"project_id": "my_program-my_project"}},
                    sort_fields=[{"submitter_id": "asc"}],
                    accessibility="accessible"
                )
        """
        if not accessibility:
            accessibility = "accessible"
        if not offset:
            offset = 0

        body = {"type": data_type, "fields": fields, "accessibility": accessibility}
        if filter_object:
            body["filter"] = filter_object
        if sort_fields:
            body["sort"] = sort_fields

        url = f"{self._auth_provider.endpoint}/guppy/download"
        response = requests.post(
            url,
            json=body,
            auth=self._auth_provider,
        )
        try:
            raise_for_status_and_print_error(response)
        except Exception:
            print(f"Unable to download.\nBody: {body}\n{response.text}")
            raise
        try:
            data = response.json()
        except Exception:
            print(f"Did not receive JSON: {response.text}")
            raise

        if offset:
            data = data[offset:]
        if first:
            data = data[:first]

        return data



import asyncio
import collections.abc
from jsonschema import Draft4Validator
import sys
import re
import requests
import random
import string
import os

from urllib.parse import urlunsplit
from urllib.parse import urlencode
from urllib.parse import urlsplit
from urllib.parse import parse_qs

from cdislogging import get_logger

logging = get_logger("__name__")

UUID_FORMAT = (
    r"^.*[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
)
MD5_FORMAT = r"^[a-fA-F0-9]{32}$"
SIZE_FORMAT = r"^[0-9]*$"
ACL_FORMAT = r"^.*$"
URL_FORMAT = r"^.*$"
AUTHZ_FORMAT = r"^.*$"


def get_random_alphanumeric(length):
    # end up with roughly the same amount of numbers as letters
    letters = string.ascii_lowercase + "".join([str(item) for item in range(0, 10)]) * 3
    return "".join(random.choice(letters) for i in range(length))


def make_folders_for_filename(filename, current_directory=None):
    """
    Make the directories up to the filename provided. Relative paths are supported.
    Ensure you supply the current directory you want relative paths to be
    based off of. Returns the absolute path for the file with the folders
    created. This does NOT create the actual file.

    Example:
        output = make_folders_for_filename(
            "../test/temp/output.txt",
            current_directory="/home/me/foobar"
        )
        print(output)
        >>> /home/me/test/temp/output.txt

    Args:
        filename (str): path to desired file
        current_directory (str, optional): current directory you want relative paths to be
    based off of

    Returns:
        str: the absolute path for the file with the folders created
    """
    current_directory = os.path.dirname(os.path.realpath(__file__))
    directory = os.path.dirname(os.path.abspath(filename))
    if not os.path.exists(directory):
        os.makedirs(directory)

    absolute_path_filename = (
        directory.rstrip("/") + "/" + os.path.abspath(filename).split("/")[-1]
    )

    return absolute_path_filename


def get_or_create_event_loop_for_thread():
    """
    Asyncio helper function to attempt to get a currently running loop and
    if there isn't one in the thread, create one and set it so future calls
    get the same event loop.
    """
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # no loop for this thread, so create one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    except AttributeError:
        # handle older versions of asyncio for previous versions of Python,
        # specifically this allows Python 3.6 asyncio to get a loop
        loop = asyncio._get_running_loop()
        if not loop:
            loop = asyncio.get_event_loop()

    return loop


def raise_for_status_and_print_error(response):
    """
    This only works for sync requests
    """
    try:
        response.raise_for_status()
    except requests.HTTPError as exception:
        print(
            "Error: status code {}; details:\n{}".format(
                response.status_code, response.text
            )
        )
        raise


def append_query_params(original_url, **kwargs):
    """
    Add additional query string arguments to the given url.

    Example call:
        new_url = append_query_params(
            original_url, error='this is an error',
            another_arg='this is another argument')
    """
    scheme, netloc, path, query_string, fragment = urlsplit(original_url)
    query_params = parse_qs(query_string)
    if kwargs is not None:
        for key, value in kwargs.items():
            query_params[key] = [value]

    new_query_string = urlencode(query_params, doseq=True)
    new_url = urlunsplit((scheme, netloc, path, new_query_string, fragment))
    return new_url


def split_url_and_query_params(url):
    """
    Given a url, return the url (no query params) and the split out the
    query paramaters separately
    """
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)
    url = urlunsplit((scheme, netloc, path, None, fragment))
    return url, query_params


def remove_trailing_whitespace_and_slashes_in_url(url):
    """
    Given a url, remove any whitespace and then slashes at the end and return url
    """
    if url:
        return url.rstrip().rstrip("/")

    return url


def _print_func_name(function):
    return "{}.{}".format(function.__module__, function.__name__)


def _print_kwargs(kwargs):
    return ", ".join("{}={}".format(k, repr(v)) for k, v in list(kwargs.items()))


def log_backoff_retry(details):
    args_str = ", ".join(map(str, details["args"]))
    kwargs_str = (
        (", " + _print_kwargs(details["kwargs"])) if details.get("kwargs") else ""
    )
    func_call_log = "{}({}{})".format(
        _print_func_name(details["target"]), args_str, kwargs_str
    )
    logging.warning(
        "backoff: call {func_call} delay {wait:0.1f} seconds after {tries} tries".format(
            func_call=func_call_log, **details
        )
    )


def log_backoff_giveup(details):
    args_str = ", ".join(map(str, details["args"]))
    kwargs_str = (
        (", " + _print_kwargs(details["kwargs"])) if details.get("kwargs") else ""
    )
    func_call_log = "{}({}{})".format(
        _print_func_name(details["target"]), args_str, kwargs_str
    )
    logging.error(
        "backoff: gave up call {func_call} after {tries} tries; exception: {exc}".format(
            func_call=func_call_log, exc=sys.exc_info(), **details
        )
    )


def log_backoff_giveup_except_on_no_retries(details):
    args_str = ", ".join(map(str, details["args"]))
    kwargs_str = (
        (", " + _print_kwargs(details["kwargs"])) if details.get("kwargs") else ""
    )
    func_call_log = "{}({}{})".format(
        _print_func_name(details["target"]), args_str, kwargs_str
    )

    if details["tries"] > 1:
        logging.error(
            "backoff: gave up call {func_call} after {tries} tries; exception: {exc}".format(
                func_call=func_call_log, exc=sys.exc_info(), **details
            )
        )


def get_delimiter_from_extension(filename):
    """
    Return the file delimter based on the extension.

    Args:
        filename (str): file name with extension

    Returns:
        str: delimeter character, either \t or ,
    """
    file_ext = os.path.splitext(filename)
    if file_ext[-1].lower() == ".tsv":
        file_delimiter = "\t"
    else:
        # default, assume CSV
        file_delimiter = ","
    return file_delimiter


def is_status_code(error, code):
    """
    Args:
        error (object): Ideally a requests.Response, this safely checks for
            known attributes where the status code might be
        code (str): The status code you want to check for in the error. ex: 404

    Returns:
        bool: Whether or not the error object contains the status code specified
    """
    code = str(code)
    return (
        str(getattr(error, "code", None)) == code
        or str(getattr(error, "status", None)) == code
        or str(getattr(error, "status_code", None)) == code
        or str(getattr(getattr(error, "response", {}), "code", "")) == code
        or str(getattr(getattr(error, "response", {}), "status", "")) == code
        or str(getattr(getattr(error, "response", {}), "status_code", "")) == code
    )


def exception_do_not_retry(error):
    if is_status_code(error, "409") or is_status_code(error, "404"):
        return True

    return False


def _verify_format(s, format):
    """
    Make sure the input is in the right format
    """
    r = re.compile(format)
    if r.match(s) is not None:
        return True
    return False


def _verify_schema(data, schema):
    validator = Draft4Validator(schema)
    validator.iter_errors(data)
    errors = [e.message for e in validator.iter_errors(data)]
    if errors:
        logging.error(
            f"Error validating package contents {data} against schema {schema}. Details: {errors}"
        )
        return False
    return True


def _standardize_str(s):
    """
    Remove unnecessary spaces

    Ex. "abc    d" -> "abc d"
    """
    memory = []
    res = ""
    for c in s:
        if c != " ":
            res += c
            memory = []
        elif not memory:
            res += c
            memory.append(" ")
    return res


def get_urls(raw_urls_string):
    """
    Given raw string like "['gs://topmed-irc-share/genomes/NWD293573%20file.b38.irc.v1.cram', 's3://nih-nhlbi-datacommons/NWD293573.b38.irc.v1.cram']"
    return a list of urls:

    [
        "gs://topmed-irc-share/genomes/NWD293573 file.b38.irc.v1.cram",
        "s3://nih-nhlbi-datacommons/NWD293573.b38.irc.v1.cram"
    ]
    """
    return [
        element.strip()
        .replace("'", "")
        .replace('"', "")
        .replace("%20", " ")
        .rstrip(",")
        for element in _standardize_str(raw_urls_string)
        .strip()
        .lstrip("[")
        .rstrip("]")
        .split(" ")
    ]


def yield_chunks(input_list, n):
    """
    Yield successive n-sized chunks from input_list.

    Args:
        input_list (list[]): arbitrary input list
        n (int): size of chunks requested

    Yields:
        list[]: chunked list
    """
    for i in range(0, len(input_list), n):
        yield input_list[i : i + n]


def deep_dict_update(a, b):
    """
    a is updated in place to include items in b

    This recursively handles nested dictionary updates
    """
    for key, value in b.items():
        if isinstance(value, collections.abc.Mapping):
            a[key] = deep_dict_update(a.get(key, {}), value)
        else:
            a[key] = value
    return a


# Default settings to control usage of backoff library.
DEFAULT_BACKOFF_SETTINGS = {
    # Disable backoff lib default logger, only show custom logs
    "logger": None,
    "on_backoff": log_backoff_retry,
    "on_giveup": log_backoff_giveup,
    "max_tries": int(os.environ.get("GEN3SDK_MAX_RETRIES", 3)),
    "giveup": exception_do_not_retry,
}

# Metadata.get settings to control usage of backoff library.
BACKOFF_NO_LOG_IF_NOT_RETRIED = {
    # Disable backoff lib default logger, only show custom logs
    "logger": None,
    "on_backoff": log_backoff_retry,
    "on_giveup": log_backoff_giveup_except_on_no_retries,
    "max_tries": int(os.environ.get("GEN3SDK_MAX_RETRIES", 3)),
    "giveup": exception_do_not_retry,
}



import itertools
import json
import requests
import os
from cdislogging import get_logger
import pandas as pd

from gen3.utils import raise_for_status_and_print_error

logging = get_logger("__name__")


class Gen3Error(Exception):
    pass


class Gen3SubmissionQueryError(Gen3Error):
    pass


class Gen3UserError(Gen3Error):
    pass


class Gen3Submission:
    """Submit/Export/Query data from a Gen3 Submission system.

    A class for interacting with the Gen3 submission services.
    Supports submitting and exporting from Sheepdog.
    Supports GraphQL queries through Peregrine.

    Args:
        auth_provider (Gen3Auth): A Gen3Auth class instance.

    Examples:
        This generates the Gen3Submission class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... sub = Gen3Submission(auth)

    """

    def __init__(self, endpoint=None, auth_provider=None):
        # auth_provider legacy interface required endpoint as 1st arg
        self._auth_provider = auth_provider or endpoint
        self._endpoint = self._auth_provider.endpoint

    def __export_file(self, filename, output):
        """Writes an API response to a file."""
        with open(filename, "w") as outfile:
            outfile.write(output)
        print("\nOutput written to file: " + filename)

    ### Program functions

    def get_programs(self):
        """List registered programs"""
        api_url = f"{self._endpoint}/api/v0/submission/"
        output = requests.get(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output.json()

    def create_program(self, json):
        """Create a program.
        Args:
            json (object): The json of the program to create

        Examples:
            This creates a program in the sandbox commons.

            >>> Gen3Submission.create_program(json)
        """
        api_url = "{}/api/v0/submission/".format(self._endpoint)
        output = requests.post(api_url, auth=self._auth_provider, json=json)
        raise_for_status_and_print_error(output)
        return output.json()

    def delete_program(self, program):
        """Delete a program.

        This deletes an empty program from the commons.

        Args:
            program (str): The program to delete.

        Examples:
            This deletes the "DCF" program.

            >>> Gen3Submission.delete_program("DCF")

        """
        api_url = "{}/api/v0/submission/{}".format(self._endpoint, program)
        output = requests.delete(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output

    ### Project functions

    def get_projects(self, program):
        """List registered projects for a given program

        Args:
            program: the name of the program you want the projects from

        Example:
            This lists all the projects under the DCF program

            >>> Gen3Submission.get_projects("DCF")

        """
        api_url = f"{self._endpoint}/api/v0/submission/{program}"
        output = requests.get(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output.json()

    def create_project(self, program, json):
        """Create a project.
        Args:
            program (str): The program to create a project on
            json (object): The json of the project to create

        Examples:
            This creates a project on the DCF program in the sandbox commons.

            >>> Gen3Submission.create_project("DCF", json)
        """
        api_url = "{}/api/v0/submission/{}".format(self._endpoint, program)
        output = requests.put(api_url, auth=self._auth_provider, json=json)
        raise_for_status_and_print_error(output)
        return output.json()

    def delete_project(self, program, project):
        """Delete a project.

        This deletes an empty project from the commons.

        Args:
            program (str): The program containing the project to delete.
            project (str): The project to delete.

        Examples:
            This deletes the "CCLE" project from the "DCF" program.

            >>> Gen3Submission.delete_project("DCF", "CCLE")

        """
        api_url = "{}/api/v0/submission/{}/{}".format(self._endpoint, program, project)
        output = requests.delete(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output

    def get_project_dictionary(self, program, project):
        """Get dictionary schema for a given project

        Args:
            program: the name of the program the project is from
            project: the name of the project you want the dictionary schema from

        Example:

            >>> Gen3Submission.get_project_dictionary("DCF", "CCLE")

        """
        api_url = f"{self._endpoint}/api/v0/submission/{program}/{project}/_dictionary"
        output = requests.get(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output.json()

    def open_project(self, program, project):
        """Mark a project ``open``. Opening a project means uploads, deletions, etc. are allowed.

        Args:
            program: the name of the program the project is from
            project: the name of the project you want to 'open'

        Example:

            >>> Gen3Submission.get_project_manifest("DCF", "CCLE")

        """
        api_url = f"{self._endpoint}/api/v0/submission/{program}/{project}/open"
        output = requests.put(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(output)
        return output.json()

    ### Record functions

    def submit_record(self, program, project, json):
        """Submit record(s) to a project as json.

        Args:
            program (str): The program to submit to.
            project (str): The project to submit to.
            json (object): The json defining the record(s) to submit. For multiple records, the json should be an array of records.

        Examples:
            This submits records to the CCLE project in the sandbox commons.

            >>> Gen3Submission.submit_record("DCF", "CCLE", json)

        """
        api_url = "{}/api/v0/submission/{}/{}".format(self._endpoint, program, project)
        logging.debug("Using the Sheepdog API URL: {}".format(api_url))

        output = requests.put(api_url, auth=self._auth_provider, json=json)
        raise_for_status_and_print_error(output)
        return output.json()

    def delete_record(self, program, project, uuid):
        """
        Delete a record from a project.

        Args:
            program (str): The program to delete from.
            project (str): The project to delete from.
            uuid (str): The uuid of the record to delete

        Examples:
            This deletes a record from the CCLE project in the sandbox commons.

            >>> Gen3Submission.delete_record("DCF", "CCLE", uuid)
        """
        return self.delete_records(program, project, [uuid])

    def delete_records(self, program, project, uuids, batch_size=100):
        """
        Delete a list of records from a project.

        Args:
            program (str): The program to delete from.
            project (str): The project to delete from.
            uuids (list): The list of uuids of the records to delete
            batch_size (int, optional, default: 100): how many records to delete at a time

        Examples:
            This deletes a list of records from the CCLE project in the sandbox commons.

            >>> Gen3Submission.delete_records("DCF", "CCLE", ["uuid1", "uuid2"])
        """
        if not uuids:
            return
        api_url = "{}/api/v0/submission/{}/{}/entities".format(
            self._endpoint, program, project
        )
        for i in itertools.count():
            uuids_to_delete = uuids[batch_size * i : batch_size * (i + 1)]
            if len(uuids_to_delete) == 0:
                break
            output = requests.delete(
                "{}/{}".format(api_url, ",".join(uuids_to_delete)),
                auth=self._auth_provider,
            )
            try:
                raise_for_status_and_print_error(output)
            except requests.exceptions.HTTPError:
                print(
                    "\n{}\nFailed to delete uuids: {}".format(
                        output.text, uuids_to_delete
                    )
                )
                raise
        return output

    def delete_node(self, program, project, node_name, batch_size=100, verbose=True):
        """
        Delete all records for a node from a project.

        Args:
            program (str): The program to delete from.
            project (str): The project to delete from.
            node_name (str): Name of the node to delete
            batch_size (int, optional, default: 100): how many records to query and delete at a time
            verbose (bool, optional, default: True): whether to print progress logs

        Examples:
            This deletes a node from the CCLE project in the sandbox commons.

            >>> Gen3Submission.delete_node("DCF", "CCLE", "demographic")
        """
        return self.delete_nodes(
            program, project, [node_name], batch_size, verbose=verbose
        )

    def delete_nodes(
        self, program, project, ordered_node_list, batch_size=100, verbose=True
    ):
        """
        Delete all records for a list of nodes from a project.

        Args:
            program (str): The program to delete from.
            project (str): The project to delete from.
            ordered_node_list (list): The list of nodes to delete, in reverse graph submission order
            batch_size (int, optional, default: 100): how many records to query and delete at a time
            verbose (bool, optional, default: True): whether to print progress logs

        Examples:
            This deletes a list of nodes from the CCLE project in the sandbox commons.

            >>> Gen3Submission.delete_nodes("DCF", "CCLE", ["demographic", "subject", "experiment"])
        """
        project_id = f"{program}-{project}"
        for node in ordered_node_list:
            if verbose:
                print(node, end="", flush=True)
            first_uuid = ""
            while True:
                query_string = f"""{{
                    {node} (first: {batch_size}, project_id: "{project_id}") {{
                        id
                    }}
                }}"""
                res = self.query(query_string)
                uuids = [x["id"] for x in res["data"][node]]
                if len(uuids) == 0:
                    break  # all done
                if first_uuid == uuids[0]:
                    raise Exception("Failed to delete. Exiting")
                first_uuid = uuids[0]
                if verbose:
                    print(".", end="", flush=True)
                self.delete_records(program, project, uuids, batch_size)
            if verbose:
                print()

    def export_record(self, program, project, uuid, fileformat, filename=None):
        """Export a single record into json.

        Args:
            program (str): The program the record is under.
            project (str): The project the record is under.
            uuid (str): The UUID of the record to export.
            fileformat (str): Export data as either 'json' or 'tsv'
            filename (str): Name of the file to export to; if no filename is provided, prints data to screen

        Examples:
            This exports a single record from the sandbox commons.

            >>> Gen3Submission.export_record("DCF", "CCLE", "d70b41b9-6f90-4714-8420-e043ab8b77b9", "json", filename="DCF-CCLE_one_record.json")

        """
        assert fileformat in [
            "json",
            "tsv",
        ], "File format must be either 'json' or 'tsv'"
        api_url = "{}/api/v0/submission/{}/{}/export?ids={}&format={}".format(
            self._endpoint, program, project, uuid, fileformat
        )
        output = requests.get(api_url, auth=self._auth_provider).text
        if filename is None:
            if fileformat == "json":
                try:
                    output = json.loads(output)
                except ValueError as e:
                    print(f"Output: {output}\nUnable to parse JSON: {e}")
                    raise
            return output
        else:
            self.__export_file(filename, output)
            return output

    def export_node(self, program, project, node_type, fileformat, filename=None):
        """Export all records in a single node type of a project.

        Args:
            program (str): The program to which records belong.
            project (str): The project to which records belong.
            node_type (str): The name of the node to export.
            fileformat (str): Export data as either 'json' or 'tsv'
            filename (str): Name of the file to export to; if no filename is provided, prints data to screen

        Examples:
            This exports all records in the "sample" node from the CCLE project in the sandbox commons.

            >>> Gen3Submission.export_node("DCF", "CCLE", "sample", "tsv", filename="DCF-CCLE_sample_node.tsv")

        """
        assert fileformat in [
            "json",
            "tsv",
        ], "File format must be either 'json' or 'tsv'"
        api_url = "{}/api/v0/submission/{}/{}/export/?node_label={}&format={}".format(
            self._endpoint, program, project, node_type, fileformat
        )
        output = requests.get(api_url, auth=self._auth_provider).text
        if filename is None:
            if fileformat == "json":
                try:
                    output = json.loads(output)
                except ValueError as e:
                    print(f"Output: {output}\nUnable to parse JSON: {e}")
                    raise
            return output
        else:
            self.__export_file(filename, output)
            return output

    ### Query functions

    def query(self, query_txt, variables=None, max_tries=1):
        """Execute a GraphQL query against a Data Commons.

        Args:
            query_txt (str): Query text.
            variables (:obj:`object`, optional): Dictionary of variables to pass with the query.
            max_tries (:obj:`int`, optional): Number of times to retry if the request fails.

        Examples:
            This executes a query to get the list of all the project codes for all the projects
            in the Data Commons.

            >>> query = "{ project(first:0) { code } }"
            ... Gen3Submission.query(query)

        """
        api_url = "{}/api/v0/submission/graphql".format(self._endpoint)
        if variables == None:
            query = {"query": query_txt}
        else:
            query = {"query": query_txt, "variables": variables}

        tries = 0
        while tries < max_tries:
            output = requests.post(api_url, auth=self._auth_provider, json=query).text
            data = json.loads(output)

            if "errors" in data:
                raise Gen3SubmissionQueryError(data["errors"])

            if not "data" in data:
                print(query_txt)
                print(data)

            tries += 1

        return data

    def get_graphql_schema(self):
        """Returns the GraphQL schema for a commons.

        This runs the GraphQL introspection query against a commons and returns the results.

        Examples:
            This returns the GraphQL schema.

            >>> Gen3Submission.get_graphql_schema()

        """
        api_url = "{}/api/v0/submission/getschema".format(self._endpoint)
        output = requests.get(api_url).text
        data = json.loads(output)
        return data

    ### Dictionary functions

    def get_dictionary_node(self, node_type):
        """Returns the dictionary schema for a specific node.

        This gets the current json dictionary schema for a specific node type in a commons.

        Args:
            node_type (str): The node_type (or name of the node) to retrieve.

        Examples:
            This returns the dictionary schema the "subject" node.

            >>> Gen3Submission.get_dictionary_node("subject")

        """
        api_url = "{}/api/v0/submission/_dictionary/{}".format(
            self._endpoint, node_type
        )
        output = requests.get(api_url).text
        data = json.loads(output)
        return data

    def get_dictionary_all(self):
        """Returns the entire dictionary object for a commons.

        This gets a json of the current dictionary schema for a commons.

        Examples:
            This returns the dictionary schema for a commons.

            >>> Gen3Submission.get_dictionary_all()

        """
        return self.get_dictionary_node("_all")

    ### File functions

    def get_project_manifest(self, program, project):
        """Get a projects file manifest

        Args:
            program: the name of the program the project is from
            project: the name of the project you want the manifest from

        Example:

            >>> Gen3Submission.get_project_manifest("DCF", "CCLE")

        """
        api_url = f"{self._endpoint}/api/v0/submission/{program}/{project}/manifest"
        output = requests.get(api_url, auth=self._auth_provider)
        return output

    def submit_file(self, project_id, filename, chunk_size=30, row_offset=0):
        """Submit data in a spreadsheet file containing multiple records in rows to a Gen3 Data Commons.

        Args:
            project_id (str): The project_id to submit to.
            filename (str): The file containing data to submit. The format can be TSV, CSV or XLSX (first worksheet only for now).
            chunk_size (integer): The number of rows of data to submit for each request to the API.
            row_offset (integer): The number of rows of data to skip; '0' starts submission from the first row and submits all data.

        Examples:
            This submits a spreadsheet file containing multiple records in rows to the CCLE project in the sandbox commons.

            >>> Gen3Submission.submit_file("DCF-CCLE","data_spreadsheet.tsv")

        """
        # Read the file in as a pandas DataFrame
        f = os.path.basename(filename)
        if f.lower().endswith(".csv"):
            df = pd.read_csv(filename, header=0, sep=",", dtype=str).fillna("")
        elif f.lower().endswith(".xlsx"):
            xl = pd.ExcelFile(filename)  # load excel file
            sheet = xl.sheet_names[0]  # sheetname
            df = xl.parse(sheet)  # save sheet as dataframe
            converters = {
                col: str for col in list(df)
            }  # make sure int isn't converted to float
            df = pd.read_excel(filename, converters=converters).fillna("")  # remove nan
        elif filename.lower().endswith((".tsv", ".txt")):
            df = pd.read_csv(filename, header=0, sep="\t", dtype=str).fillna("")
        else:
            raise Gen3UserError("Please upload a file in CSV, TSV, or XLSX format.")
        df.rename(
            columns={c: c.lstrip("*") for c in df.columns}, inplace=True
        )  # remove any leading asterisks in the DataFrame column names

        # Check uniqueness of submitter_ids:
        if len(list(df.submitter_id)) != len(list(df.submitter_id.unique())):
            raise Gen3Error(
                "Warning: file contains duplicate submitter_ids. \nNote: submitter_ids must be unique within a node!"
            )

        # Chunk the file
        print("\nSubmitting {} with {} records.".format(filename, len(df)))
        program, project = project_id.split("-", 1)
        api_url = "{}/api/v0/submission/{}/{}".format(self._endpoint, program, project)
        headers = {"content-type": "text/tab-separated-values"}

        start = row_offset
        end = row_offset + chunk_size
        chunk = df[start:end]

        count = 0

        results = {
            "invalid": {},  # these are invalid records
            "other": [],  # any unhandled API responses
            "details": [],  # entire API response details
            "succeeded": [],  # list of submitter_ids that were successfully updated/created
            "responses": [],  # list of API response codes
        }

        # Start the chunking loop:
        while (start + len(chunk)) <= len(df):
            timeout = False
            valid_but_failed = []
            invalid = []
            count += 1
            print(
                "Chunk {} (chunk size: {}, submitted: {} of {})".format(
                    count,
                    chunk_size,
                    len(results["succeeded"]) + len(results["invalid"]),
                    len(df),
                )
            )

            try:
                response = requests.put(
                    api_url,
                    auth=self._auth_provider,
                    data=chunk.to_csv(sep="\t", index=False),
                    headers=headers,
                ).text
            except requests.exceptions.ConnectionError as e:
                results["details"].append(e.message)
                continue

            # Handle the API response
            if (
                "Request Timeout" in response
                or "413 Request Entity Too Large" in response
                or "Connection aborted." in response
                or "service failure - try again later" in response
            ):  # time-out, response is not valid JSON at the moment
                print("\t Reducing Chunk Size: {}".format(response))
                results["responses"].append("Reducing Chunk Size: {}".format(response))
                timeout = True

            else:
                try:
                    json_res = json.loads(response)
                except ValueError as e:
                    print(response)
                    print(str(e))
                    raise Gen3Error("Unable to parse API response as JSON!")

                if "message" in json_res and "code" not in json_res:
                    print(
                        "\t No code in the API response for Chunk {}: {}".format(
                            count, json_res.get("message")
                        )
                    )
                    print("\t {}".format(json_res.get("transactional_errors")))
                    results["responses"].append(
                        "Error Chunk {}: {}".format(count, json_res.get("message"))
                    )
                    results["other"].append(json_res.get("transactional_errors"))

                elif "code" not in json_res:
                    print("\t Unhandled API-response: {}".format(response))
                    results["responses"].append(
                        "Unhandled API response: {}".format(response)
                    )

                elif json_res["code"] == 200:  # success
                    entities = json_res.get("entities", [])
                    print("\t Succeeded: {} entities.".format(len(entities)))
                    results["responses"].append(
                        "Chunk {} Succeeded: {} entities.".format(count, len(entities))
                    )

                    for entity in entities:
                        sid = entity["unique_keys"][0]["submitter_id"]
                        results["succeeded"].append(sid)

                elif json_res["code"] == 500:  # internal server error
                    print("\t Internal Server Error: {}".format(response))
                    results["responses"].append(
                        "Internal Server Error: {}".format(response)
                    )

                else:  # failure (400, 401, 403, 404...)
                    entities = json_res.get("entities", [])
                    print(
                        "\tChunk Failed (status code {}): {} entities.".format(
                            json_res.get("code"), len(entities)
                        )
                    )
                    results["responses"].append(
                        "Chunk {} Failed: {} entities.".format(count, len(entities))
                    )

                    for entity in entities:
                        sid = entity["unique_keys"][0]["submitter_id"]
                        if entity["valid"]:  # valid but failed
                            valid_but_failed.append(sid)
                        else:  # invalid and failed
                            message = str(entity["errors"])
                            results["invalid"][sid] = message
                            invalid.append(sid)
                    print("\tInvalid records in this chunk: {}".format(len(invalid)))

            if (
                len(valid_but_failed) > 0 and len(invalid) > 0
            ):  # if valid entities failed bc grouped with invalid, retry submission
                chunk = chunk.loc[
                    df["submitter_id"].isin(valid_but_failed)
                ]  # these are records that weren't successful because they were part of a chunk that failed, but are valid and can be resubmitted without changes
                print(
                    "Retrying submission of valid entities from failed chunk: {} valid entities.".format(
                        len(chunk)
                    )
                )

            elif (
                len(valid_but_failed) > 0 and len(invalid) == 0
            ):  # if all entities are valid but submission still failed, probably due to duplicate submitter_ids. Can remove this section once the API response is fixed: https://ctds-planx.atlassian.net/browse/PXP-3065
                raise Gen3Error(
                    "Please check your data for correct file encoding, special characters, or duplicate submitter_ids or ids."
                )

            elif timeout is False:  # get new chunk if didn't timeout
                start += chunk_size
                end = start + chunk_size
                chunk = df[start:end]

            else:  # if timeout, reduce chunk size and retry smaller chunk
                if chunk_size >= 2:
                    chunk_size = int(chunk_size / 2)
                    end = start + chunk_size
                    chunk = df[start:end]
                    print(
                        "Retrying Chunk with reduced chunk_size: {}".format(chunk_size)
                    )
                    timeout = False
                else:
                    print("Last chunk:\n{}".format(chunk))
                    raise Gen3Error(
                        "Submission is timing out. Please contact the Helpdesk."
                    )

        print("Finished data submission.")
        print("Successful records: {}".format(len(set(results["succeeded"]))))
        print("Failed invalid records: {}".format(len(results["invalid"])))

        return results



"""
The format of config file is described as following

[profile1]
key_id=key_id_example_1
api_key=api_key_example_1
access_key=access_key_example_1
api_endpoint=http://localhost:8000
use_shepherd=true
min_shepherd_version=2.0.0

[profile2]
key_id=key_id_example_2
api_key=api_key_example_2
access_key=access_key_example_2
api_endpoint=http://example.com
use_shepherd=false
min_shepherd_version=

"""
import json
from os.path import expanduser
from pathlib import Path
from collections import OrderedDict
import gen3.auth as auth_tool

from cdislogging import get_logger

logging = get_logger("__name__")

CONFIG_FILE_PATH = expanduser("~/.gen3/config")


def get_profile_from_creds(profile, cred):
    with open(expanduser(cred)) as f:
        creds_from_json = json.load(f)
        credentials = OrderedDict()
        credentials["key_id"] = creds_from_json["key_id"]
        credentials["api_key"] = creds_from_json["api_key"]
        credentials["api_endpoint"] = auth_tool.endpoint_from_token(
            credentials["api_key"]
        )
        credentials["access_key"] = auth_tool.get_access_token_with_key(credentials)
        credentials["use_shepherd"] = ""
        credentials["min_shepherd_version"] = ""
    profile_line = "[" + profile + "]\n"
    new_lines = [key + "=" + value + "\n" for key, value in credentials.items()]
    new_lines.append("\n")  # Adds an empty line between two profiles.
    return profile_line, new_lines


def get_current_config_lines():
    """Read lines from the config file if exists in ~/.gen3 folder, else create new config file"""
    try:
        with open(CONFIG_FILE_PATH) as configFile:
            logging.info(f"Reading existing config file at {CONFIG_FILE_PATH}")
            return configFile.readlines()
    except FileNotFoundError:
        Path(CONFIG_FILE_PATH).touch()
        logging.info(f"Config file doesn't exist at {CONFIG_FILE_PATH}, creating one")
        return []


def update_config_lines(lines, profile_title, new_lines):
    """Update config file contents with the new profile values"""

    if profile_title in lines:
        profile_line_index = lines.index(profile_title)
        next_profile_index = len(lines)
        for i in range(profile_line_index, len(lines)):
            if lines[i][0] == "[":
                next_profile_index = i
                break
        del lines[profile_line_index:next_profile_index]

    with open(CONFIG_FILE_PATH, "a+") as configFile:
        configFile.write(profile_title)
        configFile.writelines(new_lines)



"""
Contains class for interacting with Gen3's Metadata Service.
"""
import aiohttp
import backoff
from datetime import datetime
import requests
import json
import os
from urllib.parse import urlparse

from gen3.utils import (
    append_query_params,
    DEFAULT_BACKOFF_SETTINGS,
    BACKOFF_NO_LOG_IF_NOT_RETRIED,
    _verify_schema,
)
from gen3.auth import Gen3Auth
from gen3.tools.utils import (
    RECORD_TYPE_STANDARD_KEY,
    GUID_COLUMN_NAMES,
    FILENAME_COLUMN_NAMES,
    SIZE_COLUMN_NAMES,
    MD5_COLUMN_NAMES,
    ACLS_COLUMN_NAMES,
    URLS_COLUMN_NAMES,
    AUTHZ_COLUMN_NAMES,
    PREV_GUID_COLUMN_NAMES,
)

from cdislogging import get_logger

logging = get_logger("__name__")


PACKAGE_CONTENTS_STANDARD_KEY = "package_contents"
PACKAGE_CONTENTS_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "file_name": {
                "type": "string",
            },
            "size": {
                "type": "integer",
            },
            "hashes": {
                "type": "object",
            },
        },
        "required": ["file_name"],
        "additionalProperties": True,
    },
}


class Gen3Metadata:
    """
    A class for interacting with the Gen3 Metadata services.

    Examples:
        This generates the Gen3Metadata class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... metadata = Gen3Metadata(auth)

    Attributes:
        endpoint (str): public endpoint for reading/querying metadata - only necessary if auth_provider not provided
        auth_provider (Gen3Auth): auth manager
    """

    def __init__(
        self,
        endpoint=None,
        auth_provider=None,
        service_location="mds",
        admin_endpoint_suffix="-admin",
    ):
        """
        Initialization for instance of the class to setup basic endpoint info.

        Args:
            endpoint (str): URL for a Data Commons that has metadata service deployed
            auth_provider (Gen3Auth, optional): Gen3Auth class to handle passing your
                token, required for admin endpoints
            service_location (str, optional): deployment location relative to the
                endpoint provided
        """
        # legacy interface required endpoint as 1st arg
        if endpoint and isinstance(endpoint, Gen3Auth):
            auth_provider = endpoint
            endpoint = None
        if auth_provider and isinstance(auth_provider, Gen3Auth):
            endpoint = auth_provider.endpoint
        endpoint = endpoint.strip("/")
        # if running locally, mds is deployed by itself without a location relative
        # to the commons
        if "http://localhost" in endpoint:
            service_location = ""
            admin_endpoint_suffix = ""

        if not endpoint.endswith(service_location):
            endpoint += "/" + service_location

        self.endpoint = endpoint.rstrip("/")
        self.admin_endpoint = endpoint.rstrip("/") + admin_endpoint_suffix
        self._auth_provider = auth_provider

    def is_healthy(self):
        """
        Return if is healthy or not

        Returns:
            bool: True if healthy
        """
        try:
            response = requests.get(
                self.endpoint + "/_status", auth=self._auth_provider
            )
            response.raise_for_status()
        except Exception as exc:
            logging.error(exc)
            return False

        return response.json().get("status") == "OK"

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_version(self):
        """
        Return the version

        Returns:
            str: the version
        """
        response = requests.get(self.endpoint + "/version", auth=self._auth_provider)
        response.raise_for_status()
        return response.text

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_index_key_paths(self):
        """
        List all the metadata key paths indexed in the database.

        Returns:
            List: list of metadata key paths
        """
        response = requests.get(
            self.admin_endpoint + "/metadata_index", auth=self._auth_provider
        )
        response.raise_for_status()
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_index_key_path(self, path):
        """
        Create a metadata key path indexed in the database.

        Args:
            path (str): metadata key path
        """
        response = requests.post(
            self.admin_endpoint + f"/metadata_index/{path}", auth=self._auth_provider
        )
        response.raise_for_status()
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def delete_index_key_path(self, path):
        """
        List all the metadata key paths indexed in the database.

        Args:
            path (str): metadata key path
        """
        response = requests.delete(
            self.admin_endpoint + f"/metadata_index/{path}", auth=self._auth_provider
        )
        response.raise_for_status()
        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def query(
        self,
        query,
        return_full_metadata=False,
        limit=10,
        offset=0,
        use_agg_mds=False,
        **kwargs,
    ):
        """
        Query the metadata given a query.

        Query format is based off the logic used in the service:
            '''
            Without filters, this will return all data. Add filters as query strings like this:

            GET /metadata?a=1&b=2

        This will match all records that have metadata containing all of:
            {"a": 1, "b": 2}
            The values are always treated as strings for filtering. Nesting is supported:

            GET /metadata?a.b.c=3

        Matching records containing:
            {"a": {"b": {"c": 3}}}
            Providing the same key with more than one value filters records whose value of the given key matches any of the given values. But values of different keys must all match. For example:

            GET /metadata?a.b.c=3&a.b.c=33&a.b.d=4

        Matches these:
            {"a": {"b": {"c": 3, "d": 4}}}
            {"a": {"b": {"c": 33, "d": 4}}}
            {"a": {"b": {"c": "3", "d": 4, "e": 5}}}
            But won't match these:

            {"a": {"b": {"c": 3}}}
            {"a": {"b": {"c": 3, "d": 5}}}
            {"a": {"b": {"d": 5}}}
            {"a": {"b": {"c": "333", "d": 4}}}
            '''

        Args:
            query (str): mds query as defined by the metadata api
            return_full_metadata (bool, optional): if False will just return a list of guids
            limit (int, optional): max num records to return
            offset (int, optional): offset for output

        Returns:
            List: list of guids matching query
                OR if return_full_metadata=True
            Dict{guid: {metadata}}: Dictionary with GUIDs as keys and associated
                metadata JSON blobs as values
        """

        url = self.endpoint + f"/metadata?{query}"

        url_with_params = append_query_params(
            url, data=return_full_metadata, limit=limit, offset=offset, **kwargs
        )
        logging.debug(f"hitting: {url_with_params}")
        response = requests.get(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_get(self, guid, _ssl=None, **kwargs):
        """
        Asynchronous function to get metadata

        Args:
            guid (str): guid to use
            _ssl (None, optional): whether or not to use ssl

        Returns:
            Dict: metadata for given guid
        """
        async with aiohttp.ClientSession() as session:
            url = self.endpoint + f"/metadata/{guid}"
            url_with_params = append_query_params(url, **kwargs)

            logging.debug(f"hitting: {url_with_params}")

            async with session.get(url_with_params, ssl=_ssl) as response:
                response.raise_for_status()
                response = await response.json()

        return response

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def get(self, guid, **kwargs):
        """
        Get the metadata associated with the guid
        Args:
            guid (str): guid to use
        Returns:
            Dict: metadata for given guid
        """
        url = self.endpoint + f"/metadata/{guid}"
        url_with_params = append_query_params(url, **kwargs)
        logging.debug(f"hitting: {url_with_params}")
        response = requests.get(url_with_params, auth=self._auth_provider)

        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def batch_create(self, metadata_list, overwrite=True, **kwargs):
        """
        Create the list of metadata associated with the list of guids

        Args:
            metadata_list (List[Dict{"guid": "", "data": {}}]): list of metadata
                objects in a specific format. Expects a dict with "guid" and "data"
                fields where "data" is another JSON blob to add to the mds
            overwrite (bool, optional): whether or not to overwrite existing data
        """
        url = self.admin_endpoint + f"/metadata"

        if len(metadata_list) > 1 and (
            "guid" not in metadata_list[0] and "data" not in metadata_list[0]
        ):
            logging.warning(
                "it looks like your metadata list for bulk create is malformed. "
                "the expected format is a list of dicts that have 2 keys: 'guid' "
                "and 'data', where 'guid' is a string and 'data' is another dict. "
                f"The first element doesn't match that pattern: {metadata_list[0]}"
            )

        url_with_params = append_query_params(url, overwrite=overwrite, **kwargs)
        logging.debug(f"hitting: {url_with_params}")
        logging.debug(f"data: {metadata_list}")
        response = requests.post(
            url_with_params, json=metadata_list, auth=self._auth_provider
        )
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def create(self, guid, metadata, aliases=None, overwrite=False, **kwargs):
        """
        Create the metadata associated with the guid

        Args:
            guid (str): guid to use
            metadata (Dict): dictionary representing what will end up a JSON blob
                attached to the provided GUID as metadata
            overwrite (bool, optional): whether or not to overwrite existing data
        """
        aliases = aliases or []

        url = self.admin_endpoint + f"/metadata/{guid}"

        url_with_params = append_query_params(url, overwrite=overwrite, **kwargs)
        logging.debug(f"hitting: {url_with_params}")
        logging.debug(f"data: {metadata}")
        response = requests.post(
            url_with_params, json=metadata, auth=self._auth_provider
        )
        response.raise_for_status()

        if aliases:
            try:
                self.create_aliases(guid=guid, aliases=aliases, merge=overwrite)
            except Exception:
                logging.error(
                    "Error while attempting to create aliases: "
                    f"'{aliases}' to GUID: '{guid}' with merge={overwrite}. "
                    "GUID metadata record was created successfully and "
                    "will NOT be deleted."
                )

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_create(
        self,
        guid,
        metadata,
        aliases=None,
        overwrite=False,
        _ssl=None,
        **kwargs,
    ):
        """
        Asynchronous function to create metadata

        Args:
            guid (str): guid to use
            metadata (Dict): dictionary representing what will end up a JSON blob
                attached to the provided GUID as metadata
            overwrite (bool, optional): whether or not to overwrite existing data
            _ssl (None, optional): whether or not to use ssl
        """
        aliases = aliases or []

        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}"
            url_with_params = append_query_params(url, overwrite=overwrite, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            logging.debug(f"hitting: {url_with_params}")
            logging.debug(f"data: {metadata}")
            async with session.post(
                url_with_params, json=metadata, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()
                response = await response.json()

            if aliases:
                logging.info(f"creating aliases: {aliases}")
                try:
                    await self.async_create_aliases(
                        guid=guid, aliases=aliases, _ssl=_ssl
                    )
                except Exception:
                    logging.error(
                        "Error while attempting to create aliases: "
                        f"'{aliases}' to GUID: '{guid}'. "
                        "GUID metadata record was created successfully and "
                        "will NOT be deleted."
                    )

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def update(self, guid, metadata, aliases=None, merge=False, **kwargs):
        """
        Update the metadata associated with the guid

        Args:
            guid (str): guid to use
            metadata (Dict): dictionary representing what will end up a JSON blob
                attached to the provided GUID as metadata
        """
        aliases = aliases or []

        url = self.admin_endpoint + f"/metadata/{guid}"

        url_with_params = append_query_params(url, **kwargs)
        logging.debug(f"hitting: {url_with_params}")
        logging.debug(f"data: {metadata}")
        response = requests.put(
            url_with_params, json=metadata, auth=self._auth_provider
        )
        response.raise_for_status()

        if aliases:
            try:
                self.update_aliases(guid=guid, aliases=aliases, merge=merge)
            except Exception:
                logging.error(
                    "Error while attempting to update aliases: "
                    f"'{aliases}' to GUID: '{guid}'. "
                    "GUID metadata record was created successfully and "
                    "will NOT be deleted."
                )

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_update(
        self, guid, metadata, aliases=None, merge=False, _ssl=None, **kwargs
    ):
        """
        Asynchronous function to update metadata

        Args:
            guid (str): guid to use
            metadata (Dict): dictionary representing what will end up a JSON blob
                attached to the provided GUID as metadata
            aliases (list[str], optional): List of aliases to update the GUID with
            merge (bool, optional): Whether or not to merge metadata AND aliases
              with existing values
            _ssl (None, optional): whether or not to use ssl
            **kwargs: Description
        """
        aliases = aliases or []

        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}"
            url_with_params = append_query_params(url, merge=merge, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            async with session.put(
                url_with_params, json=metadata, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()
                response = await response.json()

            if aliases:
                try:
                    await self.async_update_aliases(
                        guid=guid, aliases=aliases, merge=merge, _ssl=_ssl
                    )
                except Exception:
                    logging.error(
                        "Error while attempting to update aliases: "
                        f"'{aliases}' to GUID: '{guid}' with merge={merge}. "
                        "GUID metadata record was created successfully and "
                        "will NOT be deleted."
                    )

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def delete(self, guid, **kwargs):
        """
        Delete the metadata associated with the guid

        Args:
            guid (str): guid to use
        """
        url = self.admin_endpoint + f"/metadata/{guid}"

        url_with_params = append_query_params(url, **kwargs)
        logging.debug(f"hitting: {url_with_params}")
        response = requests.delete(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    #
    # Alias Support
    #

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def get_aliases(self, guid, **kwargs):
        """
        Get Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to get aliases
        """
        url = self.endpoint + f"/metadata/{guid}/aliases"
        url_with_params = append_query_params(url, **kwargs)

        logging.debug(f"hitting: {url_with_params}")
        response = requests.get(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_get_aliases(self, guid, _ssl=None, **kwargs):
        """
        Asyncronously get Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to get aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.endpoint + f"/metadata/{guid}/aliases"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            logging.debug(f"hitting: {url_with_params}")
            async with session.get(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()

            return await response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def delete_alias(self, guid, alias, **kwargs):
        """
        Delete single Alias for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        url = self.admin_endpoint + f"/metadata/{guid}/aliases/{alias}"
        url_with_params = append_query_params(url, **kwargs)

        logging.debug(f"hitting: {url_with_params}")
        response = requests.delete(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_delete_alias(self, guid, alias, _ssl=None, **kwargs):
        """
        Asyncronously delete single Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}/aliases/{alias}"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            logging.debug(f"hitting: {url_with_params}")
            async with session.delete(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()

            return await response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def create_aliases(self, guid, aliases, **kwargs):
        """
        Create Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            aliases (list[str]): Aliases to set for the guid
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to create aliases
        """
        url = self.admin_endpoint + f"/metadata/{guid}/aliases"
        url_with_params = append_query_params(url, **kwargs)

        data = {"aliases": aliases}

        logging.debug(f"hitting: {url_with_params}")
        logging.debug(f"data: {data}")
        response = requests.post(url_with_params, json=data, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_create_aliases(self, guid, aliases, _ssl=None, **kwargs):
        """
        Asyncronously create Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            aliases (list[str]): Aliases to set for the guid
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to create aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}/aliases"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            data = {"aliases": aliases}

            logging.debug(f"hitting: {url_with_params}")
            logging.debug(f"data: {data}")
            async with session.post(
                url_with_params, json=data, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()
                return await response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def update_aliases(self, guid, aliases, merge=False, **kwargs):
        """
        Update Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            aliases (list[str]): Aliases to set for the guid
            merge (bool, optional): Whether or not to aliases with existing values
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to update aliases
        """
        url = self.admin_endpoint + f"/metadata/{guid}/aliases"
        url_with_params = append_query_params(url, merge=merge, **kwargs)

        data = {"aliases": aliases}

        logging.debug(f"hitting: {url_with_params}")
        logging.debug(f"data: {data}")
        response = requests.put(url_with_params, json=data, auth=self._auth_provider)
        response.raise_for_status()

        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_update_aliases(
        self, guid, aliases, merge=False, _ssl=None, **kwargs
    ):
        """
        Asyncronously update Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            aliases (list[str]): Aliases to set for the guid
            merge (bool, optional): Whether or not to aliases with existing values
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to update aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}/aliases"
            url_with_params = append_query_params(url, merge=merge, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            data = {"aliases": aliases}

            logging.debug(f"hitting: {url_with_params}")
            logging.debug(f"data: {data}")
            async with session.put(
                url_with_params, json=data, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()

                return await response.json()

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def delete_aliases(self, guid, **kwargs):
        """
        Delete all Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        url = self.admin_endpoint + f"/metadata/{guid}/aliases"
        url_with_params = append_query_params(url, **kwargs)

        logging.debug(f"hitting: {url_with_params}")
        response = requests.delete(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.text

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_delete_aliases(self, guid, _ssl=None, **kwargs):
        """
        Asyncronously delete all Aliases for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}/aliases"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            logging.debug(f"hitting: {url_with_params}")
            async with session.delete(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()

                return await response.text

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    def delete_alias(self, guid, alias, **kwargs):
        """
        Delete single Alias for the given guid

        Args:
            guid (TYPE): Globally unique ID for the metadata blob
            alias (str): alternative identifier (alias) to delete
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        url = self.admin_endpoint + f"/metadata/{guid}/aliases/{alias}"
        url_with_params = append_query_params(url, **kwargs)

        logging.debug(f"hitting: {url_with_params}")
        response = requests.delete(url_with_params, auth=self._auth_provider)
        response.raise_for_status()

        return response.text

    @backoff.on_exception(backoff.expo, Exception, **BACKOFF_NO_LOG_IF_NOT_RETRIED)
    async def async_delete_alias(self, guid, alias, _ssl=None, **kwargs):
        """
        Asyncronously delete single Aliases for the given guid

        Args:
            guid (str): Globally unique ID for the metadata blob
            alias (str): alternative identifier (alias) to delete
            _ssl (None, optional): whether or not to use ssl
            **kwargs: additional query params

        Returns:
            requests.Response: response from the request to delete aliases
        """
        async with aiohttp.ClientSession() as session:
            url = self.admin_endpoint + f"/metadata/{guid}/aliases/{alias}"
            url_with_params = append_query_params(url, **kwargs)

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self._auth_provider._get_auth_value()}

            logging.debug(f"hitting: {url_with_params}")
            async with session.delete(
                url_with_params, headers=headers, ssl=_ssl
            ) as response:
                response.raise_for_status()

                return await response.text

    def _prepare_metadata(
        self, metadata, indexd_doc, force_metadata_columns_even_if_empty
    ):
        """
        Validate and generate the provided metadata for submission to the metadata
        service.

        If the record is of type "package", also prepare package metadata.

        Args:
            metadata (dict): metadata provided by the submitter
            indexd_doc (dict): the indexd document created for this data
            force_metadata_columns_even_if_empty (bool): see description in calling function

        Returns:
            dict: metadata ready to be submitted to the metadata service
        """

        def _extract_non_indexd_metadata(metadata):
            """
            Get the "additional metadata": metadata that was provided but is
            not stored in indexd, so should be stored in the metadata service.
            """
            return {
                k: v
                for k, v in metadata.items()
                if k.lower()
                not in GUID_COLUMN_NAMES
                + FILENAME_COLUMN_NAMES
                + SIZE_COLUMN_NAMES
                + MD5_COLUMN_NAMES
                + ACLS_COLUMN_NAMES
                + URLS_COLUMN_NAMES
                + AUTHZ_COLUMN_NAMES
                + PREV_GUID_COLUMN_NAMES
            }

        to_submit = _extract_non_indexd_metadata(metadata)

        # some additional metadata columns must be validated
        valid = True

        # validate package columns
        record_type = to_submit.pop(RECORD_TYPE_STANDARD_KEY, "").strip().lower()
        package_contents = to_submit.pop(PACKAGE_CONTENTS_STANDARD_KEY, None)
        if record_type == "package":
            if package_contents:
                package_contents = json.loads(package_contents)
                if not _verify_schema(package_contents, PACKAGE_CONTENTS_SCHEMA):
                    logging.error(
                        f"ERROR: {package_contents} is not in package contents format"
                    )
                    valid = False
            # generate package metadata
            package_metadata = self._get_package_metadata(
                metadata,
                indexd_doc.file_name,
                indexd_doc.size,
                indexd_doc.hashes,
                indexd_doc.urls,
                package_contents,
            )
            to_submit.update(package_metadata)
        elif package_contents:
            logging.error(
                f"ERROR: tried to set '{PACKAGE_CONTENTS_STANDARD_KEY}' for a non-package row. Ignoring '{PACKAGE_CONTENTS_STANDARD_KEY}'. Set '{RECORD_TYPE_STANDARD_KEY}' to 'package' to create packages."
            )
            valid = False

        if not valid:
            raise Exception(f"Metadata is not valid: {metadata}")

        if not force_metadata_columns_even_if_empty:
            # remove any empty columns if we're not being forced to include them
            to_submit = {
                key: value
                for key, value in to_submit.items()
                if value is not None and value != ""
            }

        return to_submit

    def _get_package_metadata(
        self, submitted_metadata, file_name, file_size, hashes, urls, contents
    ):
        """
        The MDS Objects API currently expects files that have not been
        uploaded yet. For files we only needs to index, not upload, create
        object records manually by generating the expected object fields.
        TODO: update the MDS objects API to not create upload URLs if the
        relevant data is provided.
        """

        def _get_filename_from_urls(submitted_metadata, urls):
            file_name = ""
            if not urls:
                logging.warning(f"No URLs provided for: {submitted_metadata}")
            for url in urls:
                _file_name = os.path.basename(url)
                if not file_name:
                    file_name = _file_name
                else:
                    if file_name != _file_name:
                        logging.warning(
                            f"Received multiple URLs with different file names; will use the first URL (file name '{file_name}'): {submitted_metadata}"
                        )
            return file_name

        file_name_from_url = _get_filename_from_urls(submitted_metadata, urls)
        if not file_name:
            file_name = file_name_from_url

        now = str(datetime.utcnow())
        metadata = {
            "type": "package",
            "package": {
                "version": "0.1",
                "file_name": file_name,
                "created_time": now,
                "updated_time": now,
                "size": file_size,
                "hashes": hashes,
                "contents": contents or None,
            },
            "_upload_status": "uploaded",
        }
        return metadata



"""
gen3.external

For housing clients and wrappers against non-Gen3 APIs. At the moment, they
are usually used for gathering metadata.

If you're adding a new metadata source, use the ExternalMetadataSourceInterface.
"""
from gen3.external.external import ExternalMetadataSourceInterface



"""
Module to implement is_valid_manifest_format
"""
import warnings
import csv

from gen3.tools.utils import (
    Columns,
    MD5Validator,
    URLValidator,
    SizeValidator,
    AuthzValidator,
    EmptyWarning,
    MultiValueError,
)

from cdislogging import get_logger

logging = get_logger("__name__")


def is_valid_manifest_format(
    manifest_path,
    column_names_to_enums=None,
    allowed_protocols=["s3", "gs"],
    allow_base64_encoded_md5=False,
    error_on_empty_url=False,
    line_limit=None,
):
    """
    Validates the contents of a manifest of file objects and logs all errors
    found. Each logged error message includes a description along with the line
    number and column in which the error occurred.

    is_valid_manifest_format can validate md5, size, url and authz values by
    making use of the MD5Validator, SizeValidator, URLValidator, and
    AuthzValidator classes, respectively. See documentation in these Validator
    subclasses for more details on how specific values are validated.

    is_valid_manifest_format attempts to automatically map manifest column
    names to Validator subclasses based on the ALLOWED_COLUMN_NAMES tuple
    attribute defined in each Validator subclass. column_names_to_enums
    argument should alternatively be provided when the manifest contains custom
    column names.

    The input manifest may contain extra columns that are not intended to be
    validated, and columns can appear in any order.

    Args:
        manifest_path(str):
            path to the manifest to be validated. the manifest is assumed to be
            in a dsv (delimiter-separated values) format (e.g. csv, tsv, etc.).
            the delimiter used is automatically detected and used to parse the
            manifest
        column_names_to_enums(dict, optional):
            maps custom manifest column names to Columns enums, which
            determines which Validator subclass will be used for which manifest
            column. note that when column_names_to_enums is provided by caller,
            this means that no automatic column name matching occurs. for this
            reason, in the example below, even though it is in SizeValidator's
            ALLOWED_COLUMN_NAMES, "file_size" must still be explicitly mapped
            to Columns.SIZE. this behavior is to account for the case in which
            a user does not want validation performed on a column whose name
            would otherwise automatically be mapped to a Validator subclass
            (e.g. maybe validation on a column with the name "size" is not
            desired)

            example:
            {
                "custom md5 name": Columns.MD5,
                "file_size": Columns.SIZE,
                "custom url name": Columns.URL,
                "custom authz name": Columns.AUTHZ
            }
        allowed_protocols(list(str), optional):
            allowed protocols for url validation. note that if
            allowed_protocols is provided by caller, url values will only be
            validated using the provided protocols (e.g. if
            allowed_protocols=["http", "https"], an error would be raised when
            validating "s3://valid_bucket/valid_key" url)
        allow_base64_encoded_md5(bool, optional):
            whether or not Base64 encoded md5 values are allowed. if False,
            only hexadecimal encoded 128-bit md5 values are considered valid,
            and Base64 encoded values will be logged as errors. if True, both
            hexadecimal and Base64 encoded 128-bit md5 values are considered
            valid
        error_on_empty_url(bool, optional):
            whether to treat completely empty url values as errors

            for the following example manifest, if error_on_empty_url is False,
            a warning would be logged for the completely empty url value on
            line 2. if error_on_empty_url is True, an error would be generated
            instead
            ```
            md5,url,size
            1596f493ba9ec53023fca640fb69bd3b,,42
            ```

            note that regardless of error_on_empty_url, errors will be
            generated no matter what for arrays or quotes from which urls could
            not be extracted. for example, for the following manifest, errors
            would be generated for the url value on lines 2, 3, 4, and 5
            ```
            md5,url,size
            1596f493ba9ec53023fca640fb69bd30,"",42
            1596f493ba9ec53023fca640fb69bd31,'',43
            1596f493ba9ec53023fca640fb69bd32,[],44
            1596f493ba9ec53023fca640fb69bd33,["", ""],45
            ```
        line_limit(int, optional):
            number of lines in manifest to validate including the header. if
            not provided, every line is validated

    Returns:
        bool: True if no errors were found in manifest. False otherwise
    """
    logging.info(f'validating "{manifest_path}" manifest')
    warnings.filterwarnings("error")

    enums_to_validators = _init_enums_to_validators(
        allowed_protocols, allow_base64_encoded_md5, error_on_empty_url
    )
    with open(manifest_path, "r", encoding="utf-8-sig") as dsv_file:
        dsv_reader = _get_dsv_reader(dsv_file)
        manifest_column_names = dsv_reader.fieldnames
        manifest_column_names_to_validators = _get_manifest_column_names_to_validators(
            manifest_column_names, enums_to_validators, column_names_to_enums
        )
        _log_manifest_column_names_to_validators(manifest_column_names_to_validators)
        manifest_is_valid = _validate_manifest_column_names(
            manifest_column_names_to_validators, enums_to_validators, error_on_empty_url
        )
        if line_limit is None or line_limit > 1:
            manifest_is_valid = (
                _validate_rows(
                    dsv_reader, manifest_column_names_to_validators, line_limit
                )
                and manifest_is_valid
            )

    _log_summary(manifest_is_valid, manifest_path, line_limit)
    return manifest_is_valid


def _init_enums_to_validators(
    allowed_protocols, allow_base64_encoded_md5, error_on_empty_url
):
    """
    Initialize Validator subclass instances (i.e. validators) with provided
    arguments and map Columns enums to instantiated validators

    Args:
        allowed_protocols(list(str)): allowed protocols for URLValidator
        allow_base64_encoding(bool): whether to allow base64 encoding for
            MD5Validator
        error_on_empty_url(bool): whether to error on an empty url for
            URLValidator

    Returns:
        dict: maps Columns enums to corresponding instantiated validators
    """
    return {
        Columns.MD5: MD5Validator(allow_base64_encoding=allow_base64_encoded_md5),
        Columns.URL: URLValidator(
            allowed_protocols=allowed_protocols, error_on_empty=error_on_empty_url
        ),
        Columns.SIZE: SizeValidator(),
        Columns.AUTHZ: AuthzValidator(),
    }


def _get_dsv_reader(dsv_file):
    """
    Detect the delimiter used in opened dsv (delimiter-separated values) file
    and return csv.DictReader object initialized to iterate over file contents

    Args:
        dsv_file(file object): opened dsv file object

    Returns:
        csv.DictReader: reader object initialized to iterate over file contents
    """
    dialect = csv.Sniffer().sniff(dsv_file.readline())
    dsv_file.seek(0)
    return csv.DictReader(dsv_file, dialect=dialect, quoting=csv.QUOTE_NONE)


def _get_manifest_column_names_to_validators(
    manifest_column_names,
    enums_to_validators,
    column_names_to_enums=None,
):
    """
    Maps manifest column names to Validator subclass instances. If
    column_names_to_enums is None, a given manifest column name is mapped to a
    validator if the manifest column name exactly matches one of the
    validator's ALLOWED_COLUMN_NAMES

    Args:
        manifest_column_names(list(str)): list of the manifest's column names
        enums_to_validators(dict): maps Columns enums to Validator subclass
            instances. example:
            {
                Columns.MD5: <MD5Validator object>,
                Columns.SIZE: <SizeValidator object>,
                Columns.URL: <URLValidator object>,
                Columns.AUTHZ: <AuthzValidator object>
            }
        column_names_to_enums(dict): custom mapping between manifest column names
            and Columns enums that will be used to determine corresponding
            validators. note that no automatic matching occurs between manifest
            column names and the validators' ALLOWED_COLUMN_NAMES when
            column_names_to_enums is supplied. for this reason, in the example
            below, even though it is in SizeValidator's ALLOWED_COLUMN_NAMES,
            "size" still must be added as a key
            example:
            {
                "custom md5 name": Columns.MD5,
                "size": Columns.SIZE,
                "custom urls name": Columns.URL,
                "custom authz name": Columns.AUTHZ
            }

    Returns:
        dict: manifest column names mapped to Validator subclass instances.
            example:
            {
                "custom md5 name": <MD5Validator object>,
                "size": <SizeValidator object>,
                "custom urls name": <URLValidator object>,
                "custom authz name": <AuthzValidator object>
            }
    """
    manifest_column_names_to_validators = {}
    if column_names_to_enums is None:
        for validator in enums_to_validators.values():
            for allowed_column_name in validator.ALLOWED_COLUMN_NAMES:
                if allowed_column_name in manifest_column_names:
                    manifest_column_names_to_validators[allowed_column_name] = validator
        return manifest_column_names_to_validators

    for column_name in manifest_column_names:
        if column_name in column_names_to_enums:
            manifest_column_names_to_validators[column_name] = enums_to_validators[
                column_names_to_enums[column_name]
            ]

    return manifest_column_names_to_validators


def _validate_manifest_column_names(
    manifest_column_names_to_validators, enums_to_validators, error_on_empty_url
):
    """
    Validates that manifest has all required columns. Column names
    corresponding to Columns.MD5 and Columns.SIZE enums are required, and
    error_on_empty_url determines if Columns.URL is required

    Args:
        manifest_column_names_to_validators(dict): maps manifest column names
            to Validator subclass instances. example:
            {
                "md5": <MD5Validator object>,
                "size": <SizeValidator object>,
                "urls": <URLValidator object>,
                "authz": <AuthzValidator object>
            }
        enums_to_validators(dict): maps Columns enums to Validator subclass
            instances. example:
            {
                Columns.MD5: <MD5Validator object>,
                Columns.SIZE: <SizeValidator object>,
                Columns.URL: <URLValidator object>,
                Columns.AUTHZ: <AuthzValidator object>
            }
        error_on_empty_url(bool): if True, indicates that url column name is
            required in manifest

    Returns:
        bool: True if manifest has all required columns, False otherwise
    """

    def check_column(column_enum):
        """
        Helper method to determine if a manifest column name was mapped to
        column_enum

        Args:
            column_enum(Columns enum): identifies column (e.g. Columns.MD5)

        Returns:
            bool: False if a manifest column name was not mapped to
                column_enum and column_enum is required. Otherwise, True
        """
        validator = enums_to_validators[column_enum]
        if validator not in manifest_column_names_to_validators.values():
            message = f'line 1, could not find a column name corresponding to required "{column_enum}". Default accepted column names for "{column_enum}" are {validator.ALLOWED_COLUMN_NAMES}'
            if column_enum == Columns.URL and not error_on_empty_url:
                logging.warning(message)
            else:
                logging.error(message)
                return False

        return True

    return all([check_column(c) for c in [Columns.MD5, Columns.SIZE, Columns.URL]])


def _log_manifest_column_names_to_validators(manifest_column_names_to_validators):
    """
    Logs which manifest column names were mapped to which validators. Logs an
    error if no manifest column names were mapped.

    Args:
        manifest_column_names_to_validators(dict): maps manifest column names
            to instances of subclasses of Validator. example:
            {
                "md5": <MD5Validator object>,
                "size": <SizeValidator object>,
                "urls": <URLValidator object>,
                "authz": <AuthzValidator object>
            }

    Returns:
        None
    """
    if not manifest_column_names_to_validators:
        logging.error("line 1, no manifest columns were mapped to validators")
    for manifest_column_name, validator in manifest_column_names_to_validators.items():
        logging.info(
            f'mapped manifest column "{manifest_column_name}" to "{validator.__class__.__name__}" class instance'
        )


def _validate_rows(dsv_reader, manifest_column_names_to_validators, line_limit=None):
    """
    Loops over manifest rows starting from line 2, validating each row's values
    by calling validate method on the corresponding Validator subclass
    instance. Errors raised by the validate method are logged with the line
    number and column in which the error occurred.

    Args:
        dsv_reader(csv.DictReader): reader object in a state such that calling
            next on it will return the second row in the manifest
        manifest_column_names_to_validators(dict): maps manifest column names
            to instances of subclasses of Validator. no validation is performed
            for columns that do not appear as keys in
            manifest_column_names_to_validators. example:
            {
                "md5": <MD5Validator object>,
                "size": <SizeValidator object>,
                "urls": <URLValidator object>,
                "authz": <AuthzValidator object>
            }
        line_limit(int): the line number in the manifest to validate
            up to (e.g. if line_number is 4, _validate_rows validates lines
            2 through 4 and stops). if line_number is None, then _validate_rows
            validates up to the last line in the manifest

    Returns:
        bool: true if no errors were found, false otherwise
    """
    rows_are_valid = True
    for line_number, row in enumerate(dsv_reader, 2):
        row_items = row.items()
        if len(row_items) != len(dsv_reader.fieldnames):
            logging.warning(
                f"line {line_number}, number of fields ({len(row_items)}) in row is unequal to number of column names in manifest ({len(dsv_reader.fieldnames)})"
            )

        for column_name, value in row_items:
            if column_name in manifest_column_names_to_validators:
                validator = manifest_column_names_to_validators[column_name]
                try:
                    validator.validate(value)
                except EmptyWarning:
                    logging.warning(
                        f'line {line_number}, "{column_name}" field is empty'
                    )
                except MultiValueError as e:
                    rows_are_valid = False
                    logging.error(f'line {line_number}, "{column_name}" values {e}')
                except ValueError as e:
                    rows_are_valid = False
                    logging.error(f'line {line_number}, "{column_name}" value {e}')

        if line_number == line_limit:
            break

    return rows_are_valid


def _log_summary(manifest_is_valid, manifest_path, lines_validated):
    """
    Logs a short summary of validation that potentially includes number of
    lines validated and whether errors were found.

    Args:
        manifest_is_valid(bool): whether manifest is valid
        manifest_path(str): the path to the manifest that was validated
        lines_validated(int or NoneType): if an int, the number of lines in
            manifest that were validated. if is None, indicates that every
            line was validated

    Returns:
        None
    """
    summary = f'finished validating "{manifest_path}" manifest'
    if type(lines_validated) == int:
        summary = f'finished validating first {lines_validated} lines in "{manifest_path}" manifest'
        if lines_validated == 1:
            summary = f'finished validating first line in "{manifest_path}" manifest'
    if manifest_is_valid:
        summary += ", no errors were found"
    logging.info(summary)



import json
import requests
import json
import asyncio
import aiohttp
import aiofiles
import time
from tqdm import tqdm
from types import SimpleNamespace as Namespace
import os
import requests
from pathlib import Path

from cdislogging import get_logger

from gen3.index import Gen3Index
from gen3.utils import DEFAULT_BACKOFF_SETTINGS, raise_for_status_and_print_error
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

logging = get_logger("__name__")


MAX_RETRIES = 3


class Gen3File:
    """For interacting with Gen3 file management features.

    A class for interacting with the Gen3 file download services.
    Supports getting presigned urls right now.

    Args:
        auth_provider (Gen3Auth): A Gen3Auth class instance.

    Examples:
        This generates the Gen3File class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... file = Gen3File(auth)

    """

    def __init__(self, endpoint=None, auth_provider=None):
        # auth_provider legacy interface required endpoint as 1st arg
        self._auth_provider = auth_provider or endpoint
        self._endpoint = self._auth_provider.endpoint
        self.unsuccessful_downloads = []

    def get_presigned_url(self, guid, protocol=None):
        """Generates a presigned URL for a file.

        Retrieves a presigned url for a file giving access to a file for a limited time.

        Args:
            guid (str): The GUID for the object to retrieve.
            protocol (:obj:`str`, optional): The protocol to use for picking the available URL for generating the presigned URL.

        Examples:

            >>> Gen3File.get_presigned_url(query)

        """
        api_url = "{}/user/data/download/{}".format(self._endpoint, guid)
        if protocol:
            api_url += "?protocol={}".format(protocol)
        resp = requests.get(api_url, auth=self._auth_provider)
        raise_for_status_and_print_error(resp)

        try:
            return resp.json()
        except:
            return resp.text

    def delete_file(self, guid):
        """
        This method is DEPRECATED. Use delete_file_locations() instead.
        Delete all locations of a stored data file and remove its record from indexd

        Args:
            guid (str): provide a UUID for file id to delete
        Returns:
            text: requests.delete text result
        """
        print("This method is DEPRECATED. Use delete_file_locations() instead.")
        api_url = "{}/user/data/{}".format(self._endpoint, guid)
        output = requests.delete(api_url, auth=self._auth_provider).text

        return output

    def delete_file_locations(self, guid):
        """
        Delete all locations of a stored data file and remove its record from indexd

        Args:
            guid (str): provide a UUID for file id to delete
        Returns:
            requests.Response : requests.delete result
        """
        api_url = "{}/user/data/{}".format(self._endpoint, guid)
        output = requests.delete(api_url, auth=self._auth_provider)

        return output

    def upload_file(
        self, file_name, authz=None, protocol=None, expires_in=None, bucket=None
    ):
        """
        Get a presigned url for a file to upload

        Args:
            file_name (str): file_name to use for upload
            authz (list): authorization scope for the file as list of paths, optional.
            protocol (str): Storage protocol to use for upload: "s3", "az".
                If this isn't set, the default will be "s3"
            expires_in (int): Amount in seconds that the signed url will expire from datetime.utcnow().
                Be sure to use a positive integer.
                This value will also be treated as <= MAX_PRESIGNED_URL_TTL in the fence configuration.
            bucket (str): Bucket to upload to. The bucket must be configured in the Fence instance's
                `ALLOWED_DATA_UPLOAD_BUCKETS` setting. If not specified, Fence defaults to the
                `DATA_UPLOAD_BUCKET` setting.
        Returns:
            Document: json representation for the file upload
        """
        api_url = f"{self._endpoint}/user/data/upload"
        body = {}
        if protocol:
            body["protocol"] = protocol
        if authz:
            body["authz"] = authz
        if expires_in:
            body["expires_in"] = expires_in
        if file_name:
            body["file_name"] = file_name
        if bucket:
            body["bucket"] = bucket

        headers = {"Content-Type": "application/json"}
        resp = requests.post(
            api_url, auth=self._auth_provider, json=body, headers=headers
        )
        raise_for_status_and_print_error(resp)
        try:
            data = json.loads(resp.text)
        except:
            return resp.text

        return data

    def _ensure_dirpath_exists(path: Path) -> Path:
        """Utility to create a directory if missing.
        Returns the path so that the call can be inlined in another call
        Args:
            path (Path): path to create
        Returns
            path of created directory
        """
        assert path
        out_path: Path = path

        if not out_path.exists():
            out_path.mkdir(parents=True, exist_ok=True)

        return out_path

    def download_single(self, object_id, path):
        """
        Download a single file using its GUID.

        Args:
            object_id (str): The file's unique ID
            path (str): Path to store the downloaded file at
        """
        try:
            url = self.get_presigned_url(object_id)
        except Exception as e:
            logging.critical(f"Unable to get a presigned URL for download: {e}")
            return False

        response = requests.get(url["url"], stream=True)
        if response.status_code != 200:
            logging.error(f"Response code: {response.status_code}")
            if response.status_code >= 500:
                for _ in range(MAX_RETRIES):
                    logging.info("Retrying now...")
                    # NOTE could be updated with exponential backoff
                    time.sleep(1)
                    response = requests.get(url["url"], stream=True)
                    if response.status == 200:
                        break
                if response.status != 200:
                    logging.critical("Response status not 200, try again later")
                    return False
            else:
                return False

        response.raise_for_status()

        total_size_in_bytes = int(response.headers.get("content-length"))
        total_downloaded = 0

        index = Gen3Index(self._auth_provider)
        record = index.get_record(object_id)

        filename = record["file_name"]

        out_path = Gen3File._ensure_dirpath_exists(Path(path))

        with open(os.path.join(out_path, filename), "wb") as f:
            for data in response.iter_content(4096):
                total_downloaded += len(data)
                f.write(data)

        if total_size_in_bytes == total_downloaded:
            logging.info(f"File {filename} downloaded successfully")

        else:
            logging.error(f"File {filename} not downloaded successfully")
            return False

        return True

    def upload_file_to_guid(
        self, guid, file_name, protocol=None, expires_in=None, bucket=None
    ):
        """
        Get a presigned url for a file to upload to the specified existing GUID

        Args:
            file_name (str): file_name to use for upload
            protocol (str): Storage protocol to use for upload: "s3", "az".
                If this isn't set, the default will be "s3"
            expires_in (int): Amount in seconds that the signed url will expire from datetime.utcnow().
                Be sure to use a positive integer.
                This value will also be treated as <= MAX_PRESIGNED_URL_TTL in the fence configuration.
            bucket (str): Bucket to upload to. The bucket must be configured in the Fence instance's
                `ALLOWED_DATA_UPLOAD_BUCKETS` setting. If not specified, Fence defaults to the
                `DATA_UPLOAD_BUCKET` setting.
        Returns:
            Document: json representation for the file upload
        """
        url = f"{self._endpoint}/user/data/upload/{guid}"
        params = {}
        if protocol:
            params["protocol"] = protocol
        if expires_in:
            params["expires_in"] = expires_in
        if file_name:
            params["file_name"] = file_name
        if bucket:
            params["bucket"] = bucket

        url_parts = list(urlparse(url))
        query = dict(parse_qsl(url_parts[4]))
        query.update(params)
        url_parts[4] = urlencode(query)
        url = urlunparse(url_parts)

        resp = requests.get(url, auth=self._auth_provider)
        raise_for_status_and_print_error(resp)
        return resp.json()



import click
from pfb import cli as pfb_cli

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


@click.group()
def main():
    """Gen3 Command Line Interface"""
    pass


@click.group()
def pfb():
    """Commands for working with Portable Format for Biomedical Data (PFB)"""
    pass


for command in pfb_cli.main.commands:
    pfb.add_command(pfb_cli.main.get_command(ctx=None, cmd_name=command))

# load plug-ins from entry_points
for ep in entry_points().get("gen3.plugins", []):
    ep.load()
